
"""
.. module:: color.py
    :synopsis: It contains functions that perform color space transformations.
.. moduleauthor :: Kereman Alonso (alonso.kerman5@gmail.com)
"""

import numba
from numba import cuda


@cuda.jit(device=True)
def sRGB_to_linearRGB(sRGB_vector, linearRGB_vector):
    """
        This function takes a sRGB color space vector and
        gets the values of the equivalent linearRGB color space vector.
    """

    # Gamma-expansion (nonlinear to linear)
    for RGB in range(3):
        if (sRGB_vector[RGB] <= 0.04045):
            linearRGB_vector[RGB] =\
                sRGB_vector[RGB] / 12.92
        else:
            linearRGB_vector[RGB] =\
                ((sRGB_vector[RGB] + 0.055) /
                 1.055) ** 2.4


@cuda.jit(device=True)
def linearRGB_to_sRGB(linearRGB_vector, sRGB_vector):
    """
        This function takes a linear RGB color space vector and gets
        the values of the equivalent sRGB color space vector.
    """

    # Gamma-contraction (linear to nonlinear)
    for RGB in range(3):
        if (linearRGB_vector[RGB] <= 0.0031308):
            sRGB_vector[RGB] =\
                linearRGB_vector[RGB] * 12.92
        else:
            sRGB_vector[RGB] =\
                1.055 * linearRGB_vector[RGB] ** (1 / 2.4) - 0.055


@numba.jit(nopython=True)
def n_sRGB_to_linearRGB(sRGB_vector, linearRGB_vector):
    """
        This function takes a sRGB color space vector and
        gets the values of the equivalent linearRGB color space vector.
    """

    # Gamma-expansion (nonlinear to linear)
    for RGB in range(3):
        if (sRGB_vector[RGB] <= 0.04045):
            linearRGB_vector[RGB] =\
                sRGB_vector[RGB] / 12.92
        else:
            linearRGB_vector[RGB] =\
                ((sRGB_vector[RGB] + 0.055) /
                 1.055) ** 2.4


@numba.jit(nopython=True)
def n_linearRGB_to_sRGB(linearRGB_vector, sRGB_vector):
    """
        This function takes a linear RGB color space vector and gets
        the values of the equivalent sRGB color space vector.
    """

    # Gamma-contraction (linear to nonlinear)
    for RGB in range(3):
        if (linearRGB_vector[RGB] <= 0.0031308):
            sRGB_vector[RGB] =\
                linearRGB_vector[RGB] * 12.92
        else:
            sRGB_vector[RGB] =\
                1.055 * linearRGB_vector[RGB] ** (1 / 2.4) - 0.055


@cuda.jit(device=True)
def sRGB_to_XYZ(sRGB_vector, linearRGB_vector, XYZ_vector):
    """
        This function takes a sRGB color space vector and gets
        the values of the equivalent XYZ color space vector.
    """
    # Gamma-expansion (nonlinear to linear)
    for RGB in range(3):
        if (sRGB_vector[RGB] <= 0.04045):
            linearRGB_vector[RGB] =\
                sRGB_vector[RGB] / 12.92
        else:
            linearRGB_vector[RGB] =\
                ((sRGB_vector[RGB] + 0.055) /
                 1.055) ** 2.4

    # Linear transformation matrix: linear RGB to XYZ
    XYZ_vector[0] =\
        0.4124564 * linearRGB_vector[0] +\
        0.3575761 * linearRGB_vector[1] +\
        0.1804375 * linearRGB_vector[2]
    XYZ_vector[1] =\
        0.2126729 * linearRGB_vector[0] +\
        0.7151522 * linearRGB_vector[1] +\
        0.0721750 * linearRGB_vector[2]
    XYZ_vector[2] =\
        0.0193339 * linearRGB_vector[0] +\
        0.1191920 * linearRGB_vector[1] +\
        0.9503041 * linearRGB_vector[2]


@cuda.jit(device=True)
def XYZ_to_sRGB(XYZ_vector, linearRGB_vector, sRGB_vector):
    """
        This function takes a XYZ color space vector and gets
        the values of the equivalent sRGB color space vector.
    """
    # Linear transformation matrix: XYZ to linear RGB
    linearRGB_vector[0] =\
        3.2404542 * XYZ_vector[0] +\
        -1.5371385 * XYZ_vector[1] +\
        -0.4985314 * XYZ_vector[2]
    linearRGB_vector[1] =\
        -0.9692660 * XYZ_vector[0] +\
        1.8760108 * XYZ_vector[1] +\
        0.0415560 * XYZ_vector[2]
    linearRGB_vector[2] =\
        0.0556434 * XYZ_vector[0] +\
        -0.2040259 * XYZ_vector[1] +\
        1.0572252 * XYZ_vector[2]

    # Gamma-contraction (linear to nonlinear)
    for RGB in range(3):
        if (linearRGB_vector[RGB] <= 0.0031308):
            sRGB_vector[RGB] =\
                linearRGB_vector[RGB] * 12.92
        else:
            sRGB_vector[RGB] =\
                1.055 * linearRGB_vector[RGB] ** (1 / 2.4) - 0.055
