
"""
.. module:: intersections.py
    :synopsis: It contains functions that compute intersections between
     rays and objects in a scene.
.. moduleauthor :: Kereman Alonso (alonso.kerman5@gmail.com)
"""


from numba import cuda, float32, int32
import math

import modules.math_operations as mymath


@cuda.jit(device=True)
def intersection(objects, ray_v, P_closest, n_closest, distance,
                 t_medium_closest):
    """
        This function computes the intersection of a ray and the scene. It
        does this by iterating thorough all the objects in the scene,
        calculating the intersections of the ray and each object, and finding
        which intersection is nearest from the origin of the ray.
    """

    # Find the first object that the ray intersects.
    # Get the point P_closest where the intersection happens.
    # Get the normal vector n_closest to the objects surface at P_closest.
    t_closest = math.inf
    P = cuda.local.array(3, dtype=float32)
    n = cuda.local.array(3, dtype=float32)
    t_medium = cuda.local.array(1, dtype=float32)

    index_of_closest = 0
    object_i = 0
    while (object_i < len(objects)):
        if (objects[object_i, 23] == 0):
            t_medium[0] = 0
            if (objects[object_i, 0] == 1):
                t = sphere_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14],
                    ray_v[1], ray_v[0], P, n)
            elif (objects[object_i, 0] == 2):
                t = box_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14:17],
                    objects[object_i, 17:20],
                    objects[object_i, 20:23],
                    ray_v[1], ray_v[0], P, n)
            elif (objects[object_i, 0] == 3):
                t = triangle_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14:17],
                    objects[object_i, 17:20],
                    ray_v[1], ray_v[0], P, n)
            elif (objects[object_i, 0] == 4):
                t = ellipsoid_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14:17],
                    objects[object_i, 17:20],
                    objects[object_i, 20:23],
                    ray_v[1], ray_v[0], P, n)

        if (t < t_closest):
            t_closest = t
            for index in range(3):
                P_closest[index] = P[index]
            for index in range(3):
                n_closest[index] = n[index]
            index_of_closest = object_i
            t_medium_closest[0] = t_medium[0]
        object_i += 1
    if (t_closest < math.inf):
        distance[0] = t * math.sqrt(mymath.dot_product(ray_v[1], ray_v[1]))
        return index_of_closest
    return int32(- 1)


@cuda.jit(device=True)
def sphere_intersection(C, r, d0, P0, P, n):
    """
        This function computes the intersection point P between a ray and
        a sphere, and the normal n vector of the surface of the sphere at that
        point. It returns the t value of the line defined by the ray that
        corresponds to point P.
    """

    P0C = cuda.local.array(3, dtype=float32)
    for index in range(3):
        P0C[index] = P0[index] - C[index]
    a = mymath.dot_product(d0, d0)
    b = 2*mymath.dot_product(P0C, d0)
    c = mymath.dot_product(P0C, P0C) - r**2
    discriminant = b**2 - 4 * a * c
    if discriminant < 0:
        return math.inf
    else:
        t_plus = (- b + math.sqrt(discriminant)) / (2 * a)
        t_minus = (- b - math.sqrt(discriminant)) / (2 * a)
        if t_plus < 0:
            return math.inf
        elif t_minus < 0:
            for index in range(3):
                P[index] = P0[index] + t_plus * d0[index]
            for index in range(3):
                n[index] = (P[index] - C[index])
            return t_plus
        else:
            for index in range(3):
                P[index] = P0[index] + t_minus * d0[index]
            for index in range(3):
                n[index] = (P[index] - C[index])
            return t_minus


@cuda.jit(device=True)
def box_intersection(V, L, W, H, d0, P0, P, n):
    """
        This function computes the intersection point P between a ray and
        a box, and the normal n vector of the surface of the box at that point.
        It returns the t value of the line defined by the ray that
        corresponds to point P.
    """

    t_min = float32(0)
    t_max = math.inf
    side = 0
    Ci = cuda.local.array(3, dtype=float32)
    ni = cuda.local.array(3, dtype=float32)
    n_min = cuda.local.array(3, dtype=float32)
    n_max = cuda.local.array(3, dtype=float32)
    P0Ci = cuda.local.array(3, dtype=float32)
    while ((side < 6) and (t_min <= t_max)):
        if (side == 0):
            for index in range(3):
                Ci[index] = V[index]
            mymath.cross_product(L, H, ni)
        elif (side == 1):
            for index in range(3):
                Ci[index] = V[index] + W[index]
            for index in range(3):
                ni[index] = - ni[index]
        elif (side == 2):
            for index in range(3):
                Ci[index] = V[index]
            mymath.cross_product(H, W, ni)
        elif (side == 3):
            for index in range(3):
                Ci[index] = V[index] + L[index]
            for index in range(3):
                ni[index] = - ni[index]
        elif (side == 4):
            for index in range(3):
                Ci[index] = V[index]
            mymath.cross_product(W, L, ni)
        elif (side == 5):
            for index in range(3):
                Ci[index] = V[index] + H[index]
            for index in range(3):
                ni[index] = - ni[index]

        ray_vector_sense = mymath.dot_product(d0, ni)

        for index in range(3):
            P0Ci[index] = P0[index] - Ci[index]
        ray_point_sense = mymath.dot_product(P0Ci, ni)

        if ray_vector_sense < 0:
            t_int = -ray_point_sense / ray_vector_sense
            if (t_int >= t_min):
                for index in range(3):
                    n_min[index] = ni[index]
                t_min = t_int
        elif ray_vector_sense > 0:
            t_int = -ray_point_sense / ray_vector_sense
            if (t_int <= t_max):
                for index in range(3):
                    n_max[index] = ni[index]
                t_max = t_int
        elif ray_point_sense > 0:
            t_min = t_max + 1
        side += 1
    if (t_min > t_max):
        return math.inf
    else:
        if (t_min <= 1e-6):
            for index in range(3):
                P[index] = P0[index] + t_max * d0[index]
            for index in range(3):
                n[index] = n_max[index]
            return float32(t_max)
        else:
            for index in range(3):
                P[index] = P0[index] + t_min * d0[index]
            for index in range(3):
                n[index] = n_min[index]
            return float32(t_min)


@cuda.jit(device=True)
def triangle_intersection(C, a, b, d0, P0, P, n):
    """
        This function computes the intersection point P between a ray and
        a triangle, and the normal n vector of the surface of the triangle at
        that point. It returns the t value of the line defined by the ray that
        corresponds to point P.
    """

    # Get the n normal vector to the plane
    mymath.cross_product(a, b, n)
    if (mymath.dot_product(n, d0) > 0):
        for index in range(3):
            n[index] = -n[index]

    # Ray parallel to plane
    if(abs(mymath.dot_product(n, d0)) < 1e-6):
        return math.inf
    # Get the t value of the ray line at which the intersection happens
    P0C = cuda.local.array(3, dtype=float32)
    for index in range(3):
        P0C[index] = P0[index] - C[index]
    t = - mymath.dot_product(P0C, n) / mymath.dot_product(n, d0)
    # Intersection "behind" the point P0 at from wich the ray originated
    if (t < 0):
        return math.inf
    # Get the point of intersection of the ray with the plane
    for index in range(3):
        P[index] = P0[index] + t * d0[index]
    # Check if the intersection point P lies within the triangle
    a_dot_a = mymath.dot_product(a, a)
    b_dot_b = mymath.dot_product(b, b)
    a_dot_b = mymath.dot_product(a, b)
    PC = cuda.local.array(3, dtype=float32)
    for index in range(3):
        PC[index] = P[index] - C[index]
    PC_dot_a = mymath.dot_product(PC, a)
    PC_dot_b = mymath.dot_product(PC, b)
    A = 1 / (a_dot_a * b_dot_b - a_dot_b ** 2)
    alpha = A * (b_dot_b * PC_dot_a - a_dot_b * PC_dot_b)
    beta = A * (- a_dot_b * PC_dot_a + a_dot_a * PC_dot_b)
    # P inside triangle, there is an intersection
    if (alpha > 0 and beta > 0 and alpha + beta <= 1):
        return t
    # P outside the triangle, there is no intersection
    else:
        return math.inf


@cuda.jit(device=True)
def ellipsoid_intersection(C, u, v, w, d0, P0, P, n):
    """
        This function computes the intersection point P between a ray and
        a ellipsoid, and the normal n vector of the surface of the elipsoid
        at that point. It returns the t value of the line defined by the ray
        that corresponds to point P.
    """

    M = cuda.local.array((3, 3), dtype=float32)
    for i in range(3):
        M[i, 0] = u[i]
        M[i, 1] = v[i]
        M[i, 2] = w[i]
    M_inv = cuda.local.array((3, 3), dtype=float32)
    mymath.matrix_inverse(M, M_inv)

    P0C = cuda.local.array(3, dtype=float32)
    for index in range(3):
        P0C[index] = P0[index] - C[index]

    M_inv_d0 = cuda.local.array(3, dtype=float32)
    mymath.matrix_vector_mul(M_inv, d0, M_inv_d0)

    M_inv_P0C = cuda.local.array(3, dtype=float32)
    mymath.matrix_vector_mul(M_inv, P0C, M_inv_P0C)

    a = mymath.dot_product(M_inv_d0, M_inv_d0)
    b = 2 * mymath.dot_product(M_inv_P0C, M_inv_d0)
    c = mymath.dot_product(M_inv_P0C, M_inv_P0C) - 1
    discriminant = b ** 2 - 4 * a * c
    if discriminant < 0:
        return math.inf
    else:
        t_plus = (- b + math.sqrt(discriminant)) / (2 * a)
        t_minus = (- b - math.sqrt(discriminant)) / (2 * a)
        if t_plus < 0:
            return math.inf
        elif t_minus < 0:
            for index in range(3):
                P[index] = P0[index] + t_plus * d0[index]
            # Get normal
            M_inv_T = cuda.local.array((3, 3), dtype=float32)
            mymath.matrix_transpose(M_inv, M_inv_T)
            PC = cuda.local.array(3, dtype=float32)
            for index in range(3):
                PC[index] = (P[index] - C[index])
            n_prime = cuda.local.array(3, dtype=float32)
            mymath.matrix_vector_mul(M_inv, PC, n_prime)
            mymath.matrix_vector_mul(M_inv_T, n_prime, n)

            return t_plus
        else:
            for index in range(3):
                P[index] = P0[index] + t_minus * d0[index]
            # Get normal
            M_inv_T = cuda.local.array((3, 3), dtype=float32)
            mymath.matrix_transpose(M_inv, M_inv_T)
            PC = cuda.local.array(3, dtype=float32)
            for index in range(3):
                PC[index] = (P[index] - C[index])
            n_prime = cuda.local.array(3, dtype=float32)
            mymath.matrix_vector_mul(M_inv, PC, n_prime)
            mymath.matrix_vector_mul(M_inv_T, n_prime, n)

            return t_minus
