
"""
.. module:: math_operations.py
    :synopsis: It contains mathematical operators.
.. moduleauthor :: Kereman Alonso (alonso.kerman5@gmail.com)
"""

import numba
from numba import cuda


@cuda.jit(device=True)
def cross_product(v, u, W):
    """
        This function takes v and u 3D vectors and stores their cross product
        (v x u) in W.
    """
    W[0] = v[1]*u[2]-v[2]*u[1]
    W[1] = v[2]*u[0]-v[0]*u[2]
    W[2] = v[0]*u[1]-v[1]*u[0]


@cuda.jit(device=True)
def dot_product(v, u):
    """
        This function takes v and u 3D vectors and returns their dot product
        (v . u).
    """
    return v[0]*u[0] + v[1]*u[1] + v[2]*u[2]


@cuda.jit(device=True)
def matrix_transpose(M, MT):
    """
        This function computes the transpose of M.
    """
    for i in range(3):
        for j in range(3):
            MT[j, i] = M[i, j]


@cuda.jit(device=True)
def matrix_inverse(M, M_inv):
    """
        This function computes the inverse of M.
    """
    inv_det_M = 1 / (M[0, 0] * (M[1, 1] * M[2, 2] - M[1, 2] * M[2, 1]) -
                     M[0, 1] * (M[1, 0] * M[2, 2] - M[1, 2] * M[2, 0]) +
                     M[0, 2] * (M[1, 0] * M[2, 1] - M[1, 1] * M[2, 0]))

    M_inv[0, 0] = inv_det_M * (M[1, 1] * M[2, 2] - M[1, 2] * M[2, 1])
    M_inv[0, 1] = - inv_det_M * (M[0, 1] * M[2, 2] - M[0, 2] * M[2, 1])
    M_inv[0, 2] = inv_det_M * (M[0, 1] * M[1, 2] - M[0, 2] * M[1, 1])
    M_inv[1, 0] = - inv_det_M * (M[1, 0] * M[2, 2] - M[1, 2] * M[2, 0])
    M_inv[1, 1] = inv_det_M * (M[0, 0] * M[2, 2] - M[0, 2] * M[2, 0])
    M_inv[1, 2] = - inv_det_M * (M[0, 0] * M[1, 2] - M[0, 2] * M[1, 0])
    M_inv[2, 0] = inv_det_M * (M[1, 0] * M[2, 1] - M[1, 1] * M[2, 0])
    M_inv[2, 1] = - inv_det_M * (M[0, 0] * M[2, 1] - M[0, 1] * M[2, 0])
    M_inv[2, 2] = inv_det_M * (M[0, 0] * M[1, 1] - M[0, 1] * M[1, 0])


@cuda.jit(device=True)
def matrix_vector_mul(M, v, r):
    """
        This function computes the matrix vector multiplication M.v
    """
    r[0] = dot_product(M[0], v)
    r[1] = dot_product(M[1], v)
    r[2] = dot_product(M[2], v)


@numba.jit(nopython=True)
def n_matrix_vector_mul(M, v, r):
    """
        This function computes the matrix vector multiplication M.v
    """
    r[0] = n_dot_product(M[0], v)
    r[1] = n_dot_product(M[1], v)
    r[2] = n_dot_product(M[2], v)


@numba.jit(nopython=True)
def n_dot_product(v, u):
    """
        This function takes v and u 3D vectors and returns their dot product
        (v . u).
    """
    return v[0]*u[0] + v[1]*u[1] + v[2]*u[2]
