
"""
.. module:: ray_casting.py
    :synopsis: It contains functions that compute the origin point and
    direction components of different rays.
.. moduleauthor :: Kereman Alonso (alonso.kerman5@gmail.com)
"""

from numba import cuda, float32
import math

import modules.math_operations as mymath


@cuda.jit(device=True)
def eye_ray(camera, alpha, beta, medium, ray):
    """
        This function computes the parameters of a ray that passes through
        the "eye" defined by the camera and the pixel in the screen with
        coordinates alpha and beta. The parameters are stored in ray.
    """
    for i in range(3):
        ray[0, i] = camera[i] + camera[i+9]
        ray[1, i] = camera[i] + alpha * camera[i+3] + \
            beta * camera[i+6] - ray[0, i]
    ray[2, 0] = medium


@cuda.jit(device=True)
def shadow_feeler(light_point, P, ray):
    """
        This function computes the parameters of a ray that originates at point
        P and passes through light_point. The parameters are stored in ray.
    """
    for i in range(3):
        ray[0, i] = P[i]
        ray[1, i] = light_point[i] - P[i]
    D = math.sqrt(mymath.dot_product(ray[1], ray[1]))
    for i in range(3):
        ray[1, i] = ray[1, i] / D
    ray[2, 0] = 0
    return float32(D)


@cuda.jit(device=True)
def reflected_ray(P, n, d0, medium, ray):
    """
        This function computes the reflection of a ray incident on a surface.
    """
    # Get a normal with the correct sense (towards incoming ray)
    n_prime = cuda.local.array(3, dtype=float32)
    if mymath.dot_product(n, d0) < 0:
        for index in range(3):
            n_prime[index] = n[index]
    else:
        for index in range(3):
            n_prime[index] = -n[index]
    # Get the normalization factor of n_prime
    norm_n_prime = math.sqrt(mymath.dot_product(n_prime, n_prime))
    # Normalize n_prime
    for index in range(3):
        n_prime[index] = n_prime[index] / norm_n_prime
    # Calculate the origin point of the reflected ray (epsilon distance
    # away from P point in n direction to ensure it is in the correct side
    # of the space divided by the surface)
    epsilon = float32(1.0e-3)
    P_epsilon = cuda.local.array(3, dtype=float32)
    for index in range(3):
        P_epsilon[index] = P[index] + epsilon * n_prime[index]

    norm_d0 = math.sqrt(mymath.dot_product(d0, d0))
    for i in range(3):
        ray[0, i] = P_epsilon[i]
        ray[1, i] = d0[i] / norm_d0 -\
            2 * mymath.dot_product(d0, n_prime) / norm_d0 *\
            n_prime[i]
    ray[2, 0] = medium


@cuda.jit(device=True)
def transmited_ray(P, n, d0, ni, nt, medium, ray):
    """
        This function computes the refraction of a ray incident on a surface,
        according to Snell's Law of Refraction
    """

    epsilon = float32(1.0e-3)
    P_epsilon = cuda.local.array(3, dtype=float32)
    n_ratio = ni / nt
    norm_n = math.sqrt(mymath.dot_product(n, n))
    norm_d0 = math.sqrt(mymath.dot_product(d0, d0))
    # Get the dot product
    d0_dot_n = mymath.dot_product(d0, n) / (norm_n * norm_d0)
    # Depending on how the normal n is defined, t_dot_n should be equal to
    # cos(theta_t) or - cos(theta_t)
    if (d0_dot_n <= 0):
        t_dot_n = - math.sqrt(1 - n_ratio ** 2 * (1 - d0_dot_n ** 2))
        for index in range(3):
            P_epsilon[index] = P[index] - epsilon * n[index] / norm_n
    else:
        t_dot_n = math.sqrt(1 - n_ratio ** 2 * (1 - d0_dot_n ** 2))
        for index in range(3):
            P_epsilon[index] = P[index] + epsilon * n[index] / norm_n
    for i in range(3):
        ray[0, i] = P_epsilon[i]
        ray[1, i] = (t_dot_n - n_ratio * d0_dot_n) * n[i] /\
            norm_n + n_ratio * d0[i] / norm_d0
    ray[2, 0] = medium
