
"""
.. module:: renderGPU.py
    :synopsis: It contains the GPU kernel that calculates the color of each
     pixel using ray tracing techniques.

.. moduleauthor :: Kereman Alonso (alonso.kerman5@gmail.com)
"""

from numba import cuda, float32, int32, uint8
import math
from numba.cuda.random import xoroshiro128p_uniform_float32

import modules.math_operations as mymath
import modules.ray_casting as ray_casting
import modules.intersections as intersections
import modules.color as color


@cuda.jit
def compute_radiance(d_pixel_data, d_objects, d_camera, d_lights, d_ambient,
                     d_air, nx, ny, sampling, order, rng_states):
    """
        This function renders the color of a pixel.
    """

    grid_xDim, grid_yDim = cuda.gridDim.x, cuda.gridDim.y
    absolute_xIndex, absolute_yIndex = cuda.grid(2)

    # If the thread grid is bigger than the pixel matrix, the threads with
    # indexes bigger than those of any pixel don't have to execute anything
    if absolute_xIndex >= nx or absolute_yIndex >= ny:
        return

    # This for loops asign pixels to each thread. If the thread grid is
    # smaller than the pixel matrix, each thread will execute the code
    # corresponding to multiple pixels.
    for i in range(absolute_xIndex, nx + grid_xDim, grid_xDim):
        for j in range(absolute_yIndex, ny + grid_yDim, grid_yDim):
            if i < nx and j < ny:
                # Depth of ray iteration
                d = 3

                # Get the alfa and beta screen coordinates of a pixel
                horz_scaling = ny / 2.0
                vert_scaling = nx / 2.0
                alpha = (j - horz_scaling) / horz_scaling
                beta = -(i - vert_scaling) / vert_scaling

                color_total_RGB = cuda.local.array(3, dtype=float32)

                if (sampling == 0):
                    # Cast ray_v that originates in the camera and goes
                    # through the pixel into the scene
                    alpha += 0.5 / horz_scaling
                    beta += 0.5 / horz_scaling
                    ray_v = cuda.local.array((3, 3), dtype=float32)
                    medium = 0

                    ray_casting.eye_ray(d_camera, alpha, beta, medium, ray_v)

                    ni = float32(1)
                    color_total_linearRGB = cuda.local.array(3,
                                                             dtype=float32)

                    # Use iteration to compute global illumination
                    # User must calculate the number of rays that are
                    # going to be created due to iteration by using this
                    # formula:
                    # number_of_rays = int32(2 ** (d+1) - 1)
                    # The result has to be manualy entered as the value for
                    # the argument number_of_rays in ray_trace function.
                    # This is necessary to jit compile ray_trace function

                    ray_trace(d_objects, d_lights, d_ambient, d_air,
                              ray_v, ni, d, 15, color_total_linearRGB)

                    color.linearRGB_to_sRGB(color_total_linearRGB,
                                            color_total_RGB)

                elif (sampling == 1):
                    # number_of_rays argument has to be writen in manually
                    # in the code
                    uniform_supersampling(d_camera, d_objects, d_lights,
                                          d_ambient, d_air,
                                          alpha, beta, horz_scaling,
                                          vert_scaling, d,
                                          15, order, color_total_RGB)

                elif (sampling == 2):
                    # number_of_rays argument has to be writen in manually
                    # in the code
                    jittered_supersampling(d_camera, d_objects, d_lights,
                                           d_ambient, d_air,
                                           alpha, beta, horz_scaling,
                                           vert_scaling, d,
                                           15, order, rng_states,
                                           color_total_RGB)

                for RGB in range(3):
                    int_RGB = int32(color_total_RGB[RGB] * 255)
                    if (int_RGB <= 255):
                        d_pixel_data[i, j, RGB] = uint8(int_RGB)
                    else:
                        d_pixel_data[i, j, RGB] = uint8(255)


@cuda.jit(device=True)
def uniform_supersampling(camera, objects, lights, ambient, air,
                          alpha_pix, beta_pix, horz_scaling,
                          vert_scaling, d, number_of_rays,
                          order, color_total_RGB):
    """
        This function renders the color of a pixel aplying uniform
        supersampling.
    """

    ray_v = cuda.local.array((3, 3), dtype=float32)
    medium = 0
    ni = float32(1)
    num_of_samples = order ** 2

    color_linear = cuda.local.array(3, dtype=float32)
    color_total_linear = cuda.local.array(3, dtype=float32)

    for i in range(order):
        for j in range(order):
            alpha = alpha_pix + (0.5 + i) / (order * horz_scaling)
            beta = beta_pix + (0.5 + j) / (order * vert_scaling)
            ray_casting.eye_ray(camera, alpha, beta, medium, ray_v)
            ray_trace(objects, lights, ambient, air, ray_v, ni, d,
                      number_of_rays, color_linear)

            for RGB in range(3):
                color_total_linear[RGB] +=\
                    color_linear[RGB] / num_of_samples

    color.linearRGB_to_sRGB(color_total_linear, color_total_RGB)


@cuda.jit(device=True)
def jittered_supersampling(camera, objects, lights, ambient, air,
                           alpha_pix, beta_pix, horz_scaling,
                           vert_scaling, d, number_of_rays,
                           order, rng_states, color_total_RGB):
    """
        This function renders the color of a pixel aplying jittered
        supersampling.
    """

    grid_yDim = cuda.gridDim.y
    absolute_xIndex, absolute_yIndex = cuda.grid(2)
    flatend_index = grid_yDim * absolute_xIndex + absolute_yIndex

    ray_v = cuda.local.array((3, 3), dtype=float32)
    medium = 0
    ni = float32(1)
    num_of_samples = order ** 2

    color_linear = cuda.local.array(3, dtype=float32)
    color_total_linear = cuda.local.array(3, dtype=float32)

    for i in range(order):
        for j in range(order):
            alpha = alpha_pix +\
                (xoroshiro128p_uniform_float32(rng_states,
                                               flatend_index) + i) /\
                (order * horz_scaling)
            beta = beta_pix +\
                (xoroshiro128p_uniform_float32(rng_states,
                                               flatend_index) + j) /\
                (order * vert_scaling)
            ray_casting.eye_ray(camera, alpha, beta, medium, ray_v)
            ray_trace(objects, lights, ambient, air, ray_v, ni, d,
                      number_of_rays, color_linear)

            for RGB in range(3):
                color_total_linear[RGB] +=\
                    color_linear[RGB] / num_of_samples

    color.linearRGB_to_sRGB(color_total_linear, color_total_RGB)


@cuda.jit(device=True)
def ray_trace(objects, lights, ambient, air, ray_v, ni_v, d,
              number_of_rays, color_total_linear):
    """
        This function renders the color of a light ray.
    """

    # Initialize values of color_total_linear to 0
    color_total_linear[:] = 0

    P_closest = cuda.local.array(3, dtype=float32)
    n_closest = cuda.local.array(3, dtype=float32)
    distance = cuda.local.array(1, dtype=float32)
    t_medium = cuda.local.array(1, dtype=float32)
    current_ray = cuda.local.array((3, 3), dtype=float32)
    current_color = cuda.local.array(3, dtype=float32)

    A = cuda.local.array(3, dtype=float32)

    ray_refl = cuda.local.array((3, 3), dtype=float32)
    ray_trans = cuda.local.array((3, 3), dtype=float32)

    # Create matrix to store data of each ray. The row dimension of the
    # ray_data matrix has to be entered manually, and is equal to
    # number_of_rays
    ray_data = cuda.local.array((15, 14), dtype=float32)

    # The first matrix column stores the cumulative reflection/transmission
    # coefficient
    ray_data[0, 0] = 1
    # The second matrix column stores the index of the medium the ray is
    # traveling thorough
    ray_data[0, 1] = ni_v
    # The next 9 columns store the the information that defines the ray
    for index in range(9):
        ray_data[0, index+2] = ray_v[index//3, index % 3]
    # The last 3 columns store the cumulative attenuation factor
    ray_data[0, 11:14] = 1

    for ray_index in range(number_of_rays):

        # This if statement is True for rays that are accounted for in
        # ray_data but don't exist
        if (ray_data[ray_index, 0] < 1e-5):
            if (ray_index < 2 ** d - 1):
                # 'Child ray', the reflected one
                ray_data[2*ray_index+1, 0] = 0
                # 'Child ray', the transmited one
                ray_data[2*ray_index+2, 0] = 0
            continue

        # Get current ray parameters from ray_data
        for index in range(9):
            current_ray[index//3, index % 3] = ray_data[ray_index, index+2]

        # Get the index of the first object that the ray intersects.
        # Get the point P_closest where the intersection happens.
        # Get the normal vector n_closest to the objects surface at P_closest.
        # Get the distance to the object.
        index_of_closest = intersections.intersection(objects,
                                                      current_ray, P_closest,
                                                      n_closest, distance,
                                                      t_medium)
        # If ray doesn't intersect an object, all its 'child rays'
        # don't exist. To take this into account their cumulative
        # reflection/transmission coefficient is set to zero
        if (index_of_closest == - 1):
            if (ray_index < 2 ** d - 1):
                ray_data[2*ray_index+1, 0] = 0
                ray_data[2*ray_index+2, 0] = 0
            continue

        # Get nt assuming that all objects in the scene are surrounded by air,
        # even when two objects apear to be in contact.
        ni = ray_data[ray_index, 1]
        if (ni - 1 < 1e-6):
            nt = math.sqrt(objects[index_of_closest, 9] *
                           objects[index_of_closest, 10])
            for RGB in range(3):
                A[RGB] = air[2+RGB]
        else:
            nt = float32(1)
            for RGB in range(3):
                A[RGB] = objects[index_of_closest, 6+RGB]

        if (ray_index < 2 ** d - 1):
            ray_data[2*ray_index+1, 1] = ni
            ray_data[2*ray_index+2, 1] = nt

        # Update cumulative attenuation
        for RGB in range(3):
            ray_data[ray_index, 11+RGB] *= A[RGB] ** distance[0]

        # Get the reflection coefficient R and transmission coefficient T
        R = reflection_coefficient(current_ray, n_closest, ni, nt)
        T = 1 - R
        # Multiply R and T by the specular reflection coefficient
        R = R * objects[index_of_closest, 4]
        T = T * objects[index_of_closest, 4]

        # Store the cumulative transmission/reflection coefficients of the
        # 'children rays'. Store the cumulative attenuation factors of the
        # 'children rays'.

        if (ray_index < 2 ** d - 1):
            ray_data[2*ray_index+1, 0] = R * ray_data[ray_index, 0]
            ray_data[2*ray_index+2, 0] = T * ray_data[ray_index, 0]

            for RGB in range(3):
                ray_data[2*ray_index+1, 11+RGB] = ray_data[ray_index, 11+RGB]
                ray_data[2*ray_index+2, 11+RGB] = ray_data[ray_index, 11+RGB]

        # If ray_v has intersected with an object, compute local radiance
        # component
        local_illumination(objects, lights, ambient, current_ray[1],
                           P_closest, n_closest, index_of_closest,
                           current_color)

        # Multiply cumulative attenuation factor with current color
        for RGB in range(3):
            current_color[RGB] *= ray_data[ray_index, 11+RGB]

        # Add current color to total color
        for RGB in range(3):
            color_total_linear[RGB] +=\
                ray_data[ray_index, 0] * current_color[RGB]

        if (ray_index < 2 ** d - 1):
            # Reflected ray radiance component
            if (R > 1e-5):
                r_medium = current_ray[2, 0]
                ray_casting.reflected_ray(P_closest, n_closest, current_ray[1],
                                          r_medium, ray_refl)
                for index in range(9):
                    ray_data[2*ray_index+1, index+2] = ray_refl[
                        index//3, index % 3]

            # Transmited ray radiance component
            if (T > 1e-5):
                ray_casting.transmited_ray(P_closest, n_closest,
                                           current_ray[1], ni, nt,
                                           t_medium[0], ray_trans)
                for index in range(9):
                    ray_data[2*ray_index+2, index+2] = ray_trans[
                        index//3, index % 3]


@cuda.jit(device=True)
def local_illumination(objects, lights, ambient, d0, P_closest,
                       n_closest, index_of_closest, total_radiance_RGB):
    """
        This function calculates the diffuse, specular and ambient light
        components and ads them to get the local illumination at point
        P_closest.
    """

    # Initialize values of total_radiance_RGB to 0
    total_radiance_RGB[:] = 0

    # If ray_v has intersected with an object, cast shadow feelers from
    # P_closest to each light source. If a light source is reached, add the
    # radiance that reaches the eye through the pixel, and color the pixel.
    P = cuda.local.array(3, dtype=float32)
    n = cuda.local.array(3, dtype=float32)

    linearRGB = cuda.local.array(3, dtype=float32)

    norm_n = math.sqrt(mymath.dot_product(n_closest, n_closest))
    epsilon = float32(1.0e-3)
    P_epsilon = cuda.local.array(3, dtype=float32)
    for index in range(3):
        P_epsilon[index] = P_closest[index] + epsilon * n_closest[index] /\
            norm_n

    ray_l = cuda.local.array((3, 3), dtype=float32)
    ray_l_minus = cuda.local.array(3, dtype=float32)
    ray_rl = cuda.local.array((3, 3), dtype=float32)

    for light in range(len(lights)):
        D = ray_casting.shadow_feeler(
            lights[light, 0:3], P_epsilon, ray_l)
        shadow = False

        object_i = 0
        while (object_i < len(objects)):
            if (objects[object_i, 23] == 0):
                if (objects[object_i, 0] == 1):
                    t = intersections.sphere_intersection(
                        objects[object_i, 11:14],
                        objects[object_i, 14],
                        ray_l[1], ray_l[0], P, n)
                elif (objects[object_i, 0] == 2):
                    t = intersections.box_intersection(
                        objects[object_i, 11:14],
                        objects[object_i, 14:17],
                        objects[object_i, 17:20],
                        objects[object_i, 20:23],
                        ray_l[1], ray_l[0], P, n)
                elif (objects[object_i, 0] == 3):
                    t = intersections.triangle_intersection(
                        objects[object_i, 11:14],
                        objects[object_i, 14:17],
                        objects[object_i, 17:20],
                        ray_l[1], ray_l[0], P, n)
                elif (objects[object_i, 0] == 4):
                    t = intersections.ellipsoid_intersection(
                        objects[object_i, 11:14],
                        objects[object_i, 14:17],
                        objects[object_i, 17:20],
                        objects[object_i, 20:23],
                        ray_l[1], ray_l[0], P, n)

            if (t < D):
                shadow = True
                break
            object_i += 1

        if (shadow is False):
            # Diffuse reflection component
            nl = abs(mymath.dot_product(n_closest, ray_l[1]) / norm_n)
            for RGB in range(3):
                linearRGB[RGB] = objects[
                    index_of_closest, RGB + 1]\
                    * nl * lights[light, RGB + 3]
            # # Add colors in linear RGB
            for RGB in range(3):
                total_radiance_RGB[RGB] += linearRGB[RGB]

            # Specular reflection component
            for RGB in range(3):
                ray_l_minus[RGB] = -ray_l[1, RGB]
            ray_casting.reflected_ray(P_closest, n_closest,  ray_l_minus,
                                      ray_l[2, 0], ray_rl)
            rl_dot_d0 = mymath.dot_product(ray_rl[1], d0)
            if (rl_dot_d0 < 0):
                norm_d0 = math.sqrt(mymath.dot_product(d0, d0))
                rl_dot_d0 = rl_dot_d0 / norm_d0
                Sc = objects[index_of_closest, 4]
                Se = objects[index_of_closest, 5]
                for RGB in range(3):
                    linearRGB[RGB] = Sc * abs(rl_dot_d0) ** Se *\
                        lights[light, RGB + 3]
                # # Add colors in linear RGB
                for RGB in range(3):
                    total_radiance_RGB[RGB] += linearRGB[RGB]

            # Ambient light component
            for RGB in range(3):
                linearRGB[RGB] = objects[
                    index_of_closest, RGB + 1] * ambient[RGB]
            # # Add colors in linear RGB
            for RGB in range(3):
                total_radiance_RGB[RGB] += linearRGB[RGB]


@cuda.jit(device=True)
def reflection_coefficient(ray, n, ni, nt):
    """
        This function calculates the reflection coefficient of a surface by
        applying Fresnel equations. The ratio of the magnetic permeabilities
        of the two mediums is considered to be 1.
    """

    if (nt >= math.inf):
        return float32(1)
    n_ratio = ni / nt
    # Get cos(theta_i). cos(theta_i) should always be positive, if the normal
    # vector is chosen correctly. To take into acount that the given
    # normal vector n might be the inverse of what it should be, the absolute
    # value is computed.
    cos_theta_i = abs(mymath.dot_product(ray[1], n) /
                      (math.sqrt(mymath.dot_product(ray[1], ray[1])) *
                       math.sqrt(mymath.dot_product(n, n))))

    # Check if total internal reflaction happens. If it does return R = 1,
    # if not, get the value of cos(theta_t).
    discriminator = 1 - n_ratio ** 2 * (1 - cos_theta_i ** 2)
    if (discriminator <= 0):
        return float32(1)
    cos_theta_t = math.sqrt(discriminator)

    # For light polarized in a direction perpendicular to the plane of
    # incidence, get the ratio between the amplitude of the incident light
    # wave's electric field and the amplitude of the reflected light wave's
    # electric field.
    r_perp = (n_ratio * cos_theta_i - cos_theta_t) /\
        (n_ratio * cos_theta_i + cos_theta_t)

    # Get the same ratio for light polarized in a perpendicular direction to
    # the plane
    r_parallel = (cos_theta_i - n_ratio * cos_theta_t) /\
        (cos_theta_i + n_ratio * cos_theta_t)

    # Get the reflection coefficient R
    return float32(1.0 / 2 * (r_perp ** 2 + r_parallel ** 2))
