
"""
.. module:: scene.py
    :synopsis: It contains the functions that load the scene and renders a
    scene by calling a render function that applies ray tracing techniques.
.. moduleauthor :: Kereman Alonso (alonso.kerman5@gmail.com)
"""

import numpy as np
import cupy as cp
from numba import cuda
from numba import int32
from math import cos, sin, pi
from numba.cuda.random import create_xoroshiro128p_states

import modules.renderGPU as renderGPU
import modules.math_operations as mymath
import modules.color as color


def scene_parser(filename):
    """
        This function reads a text file and stores its data in arrays.
    """
    objects = np.zeros((1, 24), dtype=np.float32)
    lights = np.zeros((1, 7), dtype=np.float32)
    camera = 0
    ambient = np.zeros(3, dtype=np.float32)
    air = np.ones(5, dtype=np.float32)

    sRGB = np.zeros(3, dtype=np.float32)
    linearRGB = np.zeros(3, dtype=np.float32)

    with open(filename, "r") as f:
        data = f.readlines()
        no_comments_data = []
        for line in data:
            if (line[0] != "#"):
                no_comments_data.extend(line.split())
        i = 0
        while (i < len(no_comments_data)):

            if (no_comments_data[i] == 'SPHERE'):
                sphere = np.zeros(24, dtype=np.float32)

                # identifier
                sphere[0] = 1
                # D
                sRGB[:] = np.array(
                    no_comments_data[i+3][1:-1].split(","),
                    dtype=np.float32)[:]
                color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                sphere[1:4] = linearRGB[:]
                # Sc
                sphere[4] = np.float32(no_comments_data[i+4])
                # Se
                sphere[5] = np.float32(no_comments_data[i+5])
                # A
                sRGB[:] = np.array(
                    no_comments_data[i+6][1:-1].split(","),
                    dtype=np.float32)[:]
                color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                sphere[6:9] = linearRGB[:]
                # epsilon
                sphere[9] = np.float32(no_comments_data[i+7])
                # mu
                sphere[10] = np.float32(no_comments_data[i+8])
                # C
                sphere[11:14] = np.array(
                    no_comments_data[i+1][1:-1].split(","),
                    dtype=np.float32)[:]
                # r
                sphere[14] = np.float32(no_comments_data[i+2])

                # OR, AND or single
                # # OR
                if (no_comments_data[i-1] != 'OR'):
                    if (no_comments_data[i+9] == 'OR'):
                        sphere[23] = 1
                else:
                    sphere[23] = 1
                # # AND
                if (no_comments_data[i-1] != 'AND'):
                    if (no_comments_data[i+9] == 'AND'):
                        sphere[23] = 2
                else:
                    sphere[23] = 2

                if (objects[0, 0] == 0):
                    objects[0] = sphere
                else:
                    objects = np.vstack((objects[:], sphere))

                i += 9

            elif (no_comments_data[i] == 'BOX'):
                box = np.zeros(24, dtype=np.float32)

                # identifier
                box[0] = 2
                # D
                sRGB[:] = np.array(
                    no_comments_data[i+5][1:-1].split(","),
                    dtype=np.float32)[:]
                color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                box[1:4] = linearRGB[:]
                # Sc
                box[4] = np.float32(no_comments_data[i+6])
                # Se
                box[5] = np.float32(no_comments_data[i+7])
                # A
                sRGB[:] = np.array(
                    no_comments_data[i+8][1:-1].split(","),
                    dtype=np.float32)[:]
                color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                box[6:9] = linearRGB[:]
                # epsilon
                box[9] = np.float32(
                    no_comments_data[i+9])
                # mu
                box[10] = np.float32(no_comments_data[i+10])
                # V
                box[11:14] = np.array(
                    no_comments_data[i+1][1:-1].split(","),
                    dtype=np.float32)[:]
                # L
                box[14:17] = np.array(
                    no_comments_data[i+2][1:-1].split(","),
                    dtype=np.float32)[:]
                # W
                box[17:20] = np.array(
                    no_comments_data[i+3][1:-1].split(","),
                    dtype=np.float32)[:]
                # H
                box[20:23] = np.array(
                    no_comments_data[i+4][1:-1].split(","),
                    dtype=np.float32)[:]

                # OR, AND or single
                # # OR
                if (no_comments_data[i-1] != 'OR'):
                    if (no_comments_data[i+11] == 'OR'):
                        box[23] = 1
                else:
                    box[23] = 1
                # # AND
                if (no_comments_data[i-1] != 'AND'):
                    if (no_comments_data[i+11] == 'AND'):
                        box[23] = 2
                else:
                    box[23] = 2

                if (objects[0, 0] == 0):
                    objects[0] = box
                else:
                    objects = np.vstack((objects[:], box))

                i += 11

            elif (no_comments_data[i] == 'POLYGON'):
                # Number of vertices
                n = np.int32(no_comments_data[i+1])
                # Relative index of surface properties
                s = n + 2
                for vertice in range(n-2):
                    triangle = np.zeros(24, dtype=np.float32)
                    # identifier
                    triangle[0] = 3
                    # D
                    sRGB[:] = np.array(
                        no_comments_data[i+s][1:-1].split(","),
                        dtype=np.float32)[:]
                    color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                    triangle[1:4] = linearRGB[:]
                    # Sc
                    triangle[4] = np.float32(no_comments_data[i+s+1])
                    # Se
                    triangle[5] = np.float32(no_comments_data[i+s+2])
                    # A
                    sRGB[:] = np.array(
                        no_comments_data[i+s+3][1:-1].split(","),
                        dtype=np.float32)[:]
                    color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                    triangle[6:9] = linearRGB[:]
                    # epsilon
                    triangle[9] = np.float32(
                        no_comments_data[i+s+4])
                    # mu
                    triangle[10] = np.float32(no_comments_data[i+s+5])
                    # C
                    triangle[11:14] = np.array(
                        no_comments_data[i+2][1:-1].split(","),
                        dtype=np.float32)[:]
                    # a
                    triangle[14:17] = np.array(
                        no_comments_data[i+3+vertice][1:-1].split(","),
                        dtype=np.float32)[:]\
                        - triangle[11:14]
                    # b
                    triangle[17:20] = np.array(
                        no_comments_data[i+4+vertice][1:-1].split(","),
                        dtype=np.float32)[:]\
                        - triangle[11:14]

                    if (objects[0, 0] == 0):
                        objects[0] = triangle
                    else:
                        objects = np.vstack((objects[:], triangle))

                i += s + 6

            elif (no_comments_data[i] == 'ELLIPSOID'):
                ellipsoid = np.zeros(24, dtype=np.float32)

                # identifier
                ellipsoid[0] = 4
                # D
                sRGB[:] = np.array(
                    no_comments_data[i+5][1:-1].split(","),
                    dtype=np.float32)[:]
                color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                ellipsoid[1:4] = linearRGB[:]
                # Sc
                ellipsoid[4] = np.float32(no_comments_data[i+6])
                # Se
                ellipsoid[5] = np.float32(no_comments_data[i+7])
                # A
                sRGB[:] = np.array(
                    no_comments_data[i+8][1:-1].split(","),
                    dtype=np.float32)[:]
                color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                ellipsoid[6:9] = linearRGB[:]
                # epsilon
                ellipsoid[9] = np.float32(
                    no_comments_data[i+9])
                # mu
                ellipsoid[10] = np.float32(no_comments_data[i+10])
                # c
                ellipsoid[11:14] = np.array(
                    no_comments_data[i+1][1:-1].split(","),
                    dtype=np.float32)[:]
                # u
                ellipsoid[14:17] = np.array(
                    no_comments_data[i+2][1:-1].split(","),
                    dtype=np.float32)[:]
                # v
                ellipsoid[17:20] = np.array(
                    no_comments_data[i+3][1:-1].split(","),
                    dtype=np.float32)[:]
                # w
                ellipsoid[20:23] = np.array(
                    no_comments_data[i+4][1:-1].split(","),
                    dtype=np.float32)[:]

                # OR, AND or single
                # # OR
                if (no_comments_data[i-1] != 'OR'):
                    if (no_comments_data[i+11] == 'OR'):
                        ellipsoid[23] = 1
                else:
                    ellipsoid[23] = 1
                # # AND
                if (no_comments_data[i-1] != 'AND'):
                    if (no_comments_data[i+11] == 'AND'):
                        ellipsoid[23] = 2
                else:
                    ellipsoid[23] = 2

                if (objects[0, 0] == 0):
                    objects[0] = ellipsoid
                else:
                    objects = np.vstack((objects[:], ellipsoid))

                i += 11

            elif (no_comments_data[i] == 'MESH'):
                obj_filename = '..\\inputs_outputs\\' + '\\'.join(
                    no_comments_data[i+1].split('/'))[:]
                position = np.array(
                    no_comments_data[i+2][1:-1].split(","),
                    dtype=np.float32)
                orientation = np.array(
                    no_comments_data[i+3][1:-1].split(","),
                    dtype=np.float32)
                scale = np.float32(no_comments_data[i+4])
                surface = np.empty(10, dtype=np.float32)
                # D
                sRGB[:] = np.array(
                    no_comments_data[i+5][1:-1].split(","),
                    dtype=np.float32)[:]
                color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                surface[0:3] = linearRGB[:]
                # Sc
                surface[3] = np.float32(no_comments_data[i+6])
                # Se
                surface[4] = np.float32(no_comments_data[i+7])
                # A
                sRGB[:] = np.array(
                    no_comments_data[i+8][1:-1].split(","),
                    dtype=np.float32)[:]
                color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                surface[5:8] = linearRGB[:]
                # epsilon
                surface[8] = np.float32(
                    no_comments_data[i+9])
                # mu
                surface[9] = np.float32(no_comments_data[i+10])
                objects = mesh_parser(objects, obj_filename, position,
                                      orientation,
                                      scale, surface)

                i += 11

            elif (no_comments_data[i] == 'CAMERA'):
                camera = np.empty(12, dtype=np.float32)

                # c
                camera[0:3] = np.array(
                    no_comments_data[i+1][1:-1].split(","),
                    dtype=np.float32)[:]
                # u
                camera[3:6] = np.array(
                    no_comments_data[i+2][1:-1].split(","),
                    dtype=np.float32)[:]
                # v
                camera[6:9] = np.array(
                    no_comments_data[i+3][1:-1].split(","),
                    dtype=np.float32)[:]
                # e
                camera[9:12] = np.array(
                    no_comments_data[i+4][1:-1].split(","),
                    dtype=np.float32)[:]

                camera = camera
                i += 5

            elif (no_comments_data[i] == 'LIGHT'):
                light = np.empty(7, dtype=np.float32)

                # p
                light[0:3] = np.array(
                    no_comments_data[i+1][1:-1].split(","),
                    dtype=np.float32)[:]
                # I
                sRGB[:] = np.array(
                    no_comments_data[i+2][1:-1].split(","),
                    dtype=np.float32)[:]
                color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                light[3:6] = linearRGB[:]
                # r
                light[6] = np.float32(no_comments_data[i+3])

                if (lights[0, 0] == 0):
                    lights[0] = light
                else:
                    lights = np.vstack((lights[:], light))

                i += 4

            elif (no_comments_data[i] == 'AMBIENT'):
                # I: RGB color of ambient light
                sRGB[:] = np.array(
                    no_comments_data[i+1][1:-1].split(","),
                    dtype=np.float32)[:]
                color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                ambient[0:3] = linearRGB[:]
                i += 2

            elif (no_comments_data[i] == 'AIR'):
                # epsilon: relative electric permitivity
                air[0] = np.float32(no_comments_data[i+1])
                # mu: relative magnetic permeability
                air[1] = np.float32(no_comments_data[i+2])
                # A: RGB attenuation factors
                sRGB[:] = np.array(
                    no_comments_data[i+3][1:-1].split(","),
                    dtype=np.float32)[:]
                color.n_sRGB_to_linearRGB(sRGB, linearRGB)
                air[2:5] = linearRGB[:]
                i += 4

            # Skip the rest of the values. If .txt has the correct format
            # this condition shoud only skip 'OR' and 'AND' values.
            else:
                i += 1
    return objects, lights, camera, ambient, air


def mesh_parser(in_objects, obj_filename, position, orientation,
                scale, surface):
    """
        This function reads a .obj file and stores its mesh data into triangle
        arrays.
    """
    # Get rotation matrix from Euler angles
    orientation = orientation * pi / 180
    u = orientation[0]
    v = orientation[1]
    w = orientation[2]
    rotation_matrix = np.zeros((3, 3), dtype=np.float32)
    rotation_matrix[0, 0] = cos(v) * cos(w)
    rotation_matrix[0, 1] = sin(u) * sin(v) * cos(w) - cos(u) * sin(w)
    rotation_matrix[0, 2] = sin(u) * sin(w) + cos(u) * sin(v) * cos(w)
    rotation_matrix[1, 0] = cos(v) * sin(w)
    rotation_matrix[1, 1] = cos(u) * cos(w) + sin(u) * sin(v) * sin(w)
    rotation_matrix[1, 2] = cos(u) * sin(v) * sin(w) - sin(u) * cos(w)
    rotation_matrix[2, 0] = -sin(v)
    rotation_matrix[2, 1] = sin(u) * cos(v)
    rotation_matrix[2, 2] = cos(u) * cos(v)

    objects = np.array(in_objects, dtype=np.float32)

    with open(obj_filename, "r") as f:
        data = f.readlines()
        no_comments_data = []
        for line in data:
            if (line[0] != "#"):
                no_comments_data.extend(line.split())
        vertices = []
        i = 0
        while (i < len(no_comments_data)):
            if (no_comments_data[i] == 'v'):
                vertex = np.array(
                    no_comments_data[i+1:i+4],
                    dtype=np.float32)
                vertices.append(vertex)
                i += 4

            elif(no_comments_data[i] == 'vn'):
                i += 4

            elif(no_comments_data[i] == 'vt'):
                i += 3

            elif(no_comments_data[i] == 'f'):
                i += 1
                polygon_vertices = []
                while (no_comments_data[i] != 'f'):
                    polygon_vertices.append(
                        vertices[
                            int(no_comments_data[i].split("/")[0]) - 1])
                    i += 1
                    if (i == len(no_comments_data)):
                        break
                n = len(polygon_vertices)
                r = np.empty(3, dtype=np.float32)
                for vertex_i in range(n):
                    # Apply Euler rotation to all vertices
                    mymath.matrix_vector_mul(rotation_matrix,
                                             polygon_vertices[vertex_i], r)
                    polygon_vertices[vertex_i] = r
                    # Apply scale to all vertices
                    polygon_vertices[vertex_i] = scale *\
                        polygon_vertices[vertex_i]

                    # Add position to all vertices
                    polygon_vertices[vertex_i] =\
                        polygon_vertices[vertex_i] + position

                # Create triangles from vertices and store them
                for vertex_i in range(n-2):
                    triangle = np.zeros(24, dtype=np.float32)
                    # identifier
                    triangle[0] = 3
                    # Surface properties
                    triangle[1:11] = surface[:]
                    # C
                    triangle[11:14] = polygon_vertices[0]
                    # a
                    triangle[14:17] = polygon_vertices[vertex_i + 1]\
                        - triangle[11:14]
                    # b
                    triangle[17:20] = polygon_vertices[vertex_i + 2]\
                        - triangle[11:14]
                    if (objects[0, 0] == 0):
                        objects[0] = triangle
                    else:
                        objects = np.vstack((objects[:], triangle))

            else:
                i += 1
        return objects


def render(objects, lights, camera, ambient, air):
    """
        This function takes the scene data and renders the scene
        using ray traicing.
    """

    ny = int32(512)  # pixels in u direction
    nx = int32(ny * np.linalg.norm(camera[6:9]) /
               np.linalg.norm(camera[3:6]))  # pixels in v direction

    d_pixel_data = cp.zeros((nx, ny, 3), dtype=np.uint8)
    d_objects = cuda.to_device(objects)
    d_camera = cuda.to_device(camera)
    d_lights = cuda.to_device(lights)
    d_ambient = cuda.to_device(ambient)
    d_air = cuda.to_device(air)

    threadsPerBlock = (int32(4), int32(4))
    numBlocks = (int32(nx // threadsPerBlock[0] if
                 nx % threadsPerBlock[0] == 0
                 else nx // threadsPerBlock[0] + 1),
                 int32(ny // threadsPerBlock[1] if
                 ny % threadsPerBlock[1] == 0
                 else ny // threadsPerBlock[1] + 1))

    # Random number generation
    rng_states = create_xoroshiro128p_states(
        int(threadsPerBlock[0] * threadsPerBlock[1] * numBlocks[0] *
            numBlocks[0]), seed=1)

    # # Order of supersampling
    # order = 0

    # No supersampling
    renderGPU.compute_radiance[numBlocks, threadsPerBlock](
                d_pixel_data, d_objects, d_camera, d_lights,
                d_ambient, d_air, nx, ny, 0, 0, rng_states)

    # # Uniform supersampling
    # renderGPU.compute_radiance[numBlocks, threadsPerBlock](
    #             d_pixel_data, d_objects, d_camera, d_lights,
    #             d_ambient, d_air, nx, ny, 1, order, rng_states)

    # # Jittered supersampling
    # renderGPU.compute_radiance[numBlocks, threadsPerBlock](
    #             d_pixel_data, d_objects, d_camera, d_lights,
    #             d_ambient, d_air, nx, ny, 2, order, rng_states)

    cuda.synchronize()

    return cp.asnumpy(d_pixel_data)
