
"""
.. module:: measure_performance.py
    :synopsis: It contains the script used to measure the performance of
    the Numba ray tracer.
.. moduleauthor:: Kereman Alonso (alonso.kerman5@gmail.com)
"""


import tkinter as tk
from PIL import ImageTk, Image
import psutil
import time
import multiprocessing
import numpy as np
import nvidia_smi
import csv

import raytracing_main
import modules.scene as scene


def measure_in_parallel(delay, sample_pid):
    """
        This function measures the CPU and memory utilizations of a
        given process.
    """
    sample_process = psutil.Process(sample_pid)
    t_0 = time.perf_counter()
    with open('..\\inputs_outputs\\csv_files\\Numba_performance.csv',
              "w", newline='') as csvfile:
        csv_writer = csv.writer(csvfile, delimiter=',',
                                quotechar='"', quoting=csv.QUOTE_MINIMAL)
        while True:
            t = time.perf_counter() - t_0

            CPU_utilization = sample_process.cpu_percent(interval=delay) /\
                psutil.cpu_count()

            # print("CPU utilization: {}%".format(CPU_utilization))
            # print("Process children: {}".format(
            #     sample_process.children(recursive=False)))

            memory = sample_process.memory_info()
            RSS = memory[0] / 1024 / 1024
            VMS = memory[1] / 1024 / 1024
            # print("RSS: {}MB".format(RSS))
            # print("VMS: {}MB".format(VMS))

            nvidia_smi.nvmlInit()
            handle = nvidia_smi.nvmlDeviceGetHandleByIndex(0)
            res = nvidia_smi.nvmlDeviceGetUtilizationRates(handle)
            GPU_utilization = res.gpu
            GPU_memory = res.memory
            # print("GPU utilization: {}%".format(GPU_utilization))
            # print("GPU memory: {}%".format(GPU_memory))
            # print('')
            # sys.stdout.flush()

            csv_writer.writerow([t, CPU_utilization, RSS, VMS,
                                 GPU_utilization, GPU_memory])
            csvfile.flush()
            # time.sleep(delay)


def measure_dt():
    """
        This function computes the precision of perf_counter().
    """

    # measure the smallest time delta by spinning until the time changes
    t0 = time.perf_counter()
    t1 = t0
    while t1 == t0:
        t1 = time.perf_counter()
    return t1-t0


if __name__ == "__main__":

    t_0 = time.perf_counter()

    # Measure the smallest time delta
    average_dt = np.average(np.array([measure_dt() for i in range(int(1e3))]))

    print("Precision of time measurements:")
    print("--- {} seconds ---".format(average_dt))

    scene_file = '..\\inputs_outputs\\scenes\\comparison.txt'

    # Measure scene loading time
    lo_start_time = time.perf_counter()
    iteration_num = 1
    for index in range(iteration_num):
        # load scene from .txt file
        objects, lights, camera, ambient, air =\
            scene.scene_parser(scene_file)
    print("Numba, average scene loading time, {} iterations:".format(
        iteration_num))
    print("--- {} seconds ---".format((
        time.perf_counter() - lo_start_time) / iteration_num))

    # First call to just in time compile
    jit_start_time = time.perf_counter()
    pixel_data = scene.render(objects, lights, camera, ambient, air)
    print("JIT compilation time:")
    print("--- {} seconds ---".format(time.perf_counter() - jit_start_time))

    # Measure rendering time
    ren_start_time = time.perf_counter()
    iteration_num = 1
    for index in range(iteration_num):
        # render the scene and pass correspondin pixel values to pixel_data
        pixel_data = scene.render(objects, lights, camera, ambient, air)
    print("Numba, average rendering time, {} iterations".format(
        iteration_num))
    print("--- {} seconds ---".format((
        time.perf_counter() - ren_start_time) / iteration_num))

    # Measure image display time
    dis_start_time = time.perf_counter()

    iteration_num = 1
    for index in range(iteration_num):
        # convert pixel_data matrix into a .png image
        rendered_image = Image.fromarray(pixel_data)

        # create window and set title
        window = tk.Tk()
        window.title("TFG IE")

        # create frame inside window
        image_frame = tk.LabelFrame(window)
        image_frame.pack(padx=10, pady=10)

        # convert renderer_image into a tk compatible image and
        # display in image_frame
        image_tk = ImageTk.PhotoImage(rendered_image)
        label = tk.Label(image_frame, image=image_tk)
        label.pack()

        # Save image
        rendered_image.save('..\\inputs_outputs\\outputs\\render.png')

    print("Numba, average image displaying time, {} iterations:".format(
        iteration_num))
    print("--- {} seconds ---".format((
        time.perf_counter() - dis_start_time) / iteration_num))

    print("Total time: {}s".format(time.perf_counter() - t_0))

    # window.mainloop()

    # CPU utilization and memory measurements
    print('Measuring CPU utilization and memory...')
    # Create process that will perform the rendering
    renderer = multiprocessing.Process(
        target=raytracing_main.get_image, args=(scene_file,))
    renderer.daemon = False
    renderer.start()

    # Start parallel measurer process
    delay = 0.2
    sample_pid = renderer.pid
    measurer = multiprocessing.Process(
        target=measure_in_parallel, args=(delay, sample_pid,))
    measurer.daemon = True
    measurer.start()

    renderer.join()

    measurer.terminate()
    measurer.join()
