
"""
.. module:: intersections.py
    :synopsis: It contains functions that compute intersections between
     rays and objects in a scene.
.. moduleauthor :: Kereman Alonso (alonso.kerman5@gmail.com)
"""


import numpy as np
from numpy import float32, int32
import numba
import math

import modules.math_operations as mymath


@numba.jit(nopython=True)
def intersection(objects, ray_v, P_closest, n_closest, distance,
                 t_medium_closest):
    """
        This function computes the intersection of a ray and the scene. It
        does this by iterating thorough all the objects in the scene,
        calculating the intersections of the ray and each object, and finding
        which intersection is nearest from the origin of the ray.
    """

    # Find the first object that the ray intersects.
    # Get the point P_closest where the intersection happens.
    # Get the normal vector n_closest to the objects surface at P_closest.
    t_closest = np.Inf
    P = np.zeros(3, dtype=np.float32)
    n = np.zeros(3, dtype=np.float32)
    t_medium = np.zeros(1, dtype=np.float32)
    boolean_op_num = np.zeros(1, dtype=int32)

    index_of_closest = 0
    object_i = 0
    while (object_i < len(objects)):
        if (objects[object_i, 23] == 0):
            t_medium[0] = 0
            if (objects[object_i, 0] == 1):
                t = sphere_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14],
                    ray_v[1], ray_v[0], P, n)
            elif (objects[object_i, 0] == 2):
                t = box_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14:17],
                    objects[object_i, 17:20],
                    objects[object_i, 20:23],
                    ray_v[1], ray_v[0], P, n)
            elif (objects[object_i, 0] == 3):
                t = triangle_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14:17],
                    objects[object_i, 17:20],
                    ray_v[1], ray_v[0], P, n)
            elif (objects[object_i, 0] == 4):
                t = ellipsoid_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14:17],
                    objects[object_i, 17:20],
                    objects[object_i, 20:23],
                    ray_v[1], ray_v[0], P, n)
        elif (objects[object_i, 23] == 1):
            t = OR_intersection(object_i, objects, ray_v[1], ray_v[0],
                                ray_v[2, 0], P, n, t_medium, boolean_op_num)
            object_i += boolean_op_num[0] - 1

        elif (objects[object_i, 23] == 2):
            t = AND_intersection(object_i, objects, ray_v[1], ray_v[0],
                                 ray_v[2, 0], P, n, t_medium, boolean_op_num)
            object_i += boolean_op_num[0] - 1

        if (t < t_closest):
            t_closest = t
            for index in range(3):
                P_closest[index] = P[index]
            for index in range(3):
                n_closest[index] = n[index]
            index_of_closest = object_i
            t_medium_closest[0] = t_medium[0]
        object_i += 1
    if (t_closest < math.inf):
        distance[0] = t * math.sqrt(mymath.dot_product(ray_v[1], ray_v[1]))
        return index_of_closest
    return int32(- 1)


@numba.jit(nopython=True)
def sphere_intersection(C, r, d0, P0, P, n):
    """
        This function computes the intersection point P between a ray and
        a sphere, and the normal n vector of the surface of the sphere at that
        point. It returns the t value of the line defined by the ray that
        corresponds to point P.
    """

    P0C = np.zeros(3, dtype=np.float32)
    for index in range(3):
        P0C[index] = P0[index] - C[index]
    a = mymath.dot_product(d0, d0)
    b = 2*mymath.dot_product(P0C, d0)
    c = mymath.dot_product(P0C, P0C) - r**2
    discriminant = b**2 - 4 * a * c
    if discriminant < 0:
        return np.Inf
    else:
        t_plus = (- b + np.sqrt(discriminant)) / (2 * a)
        t_minus = (- b - np.sqrt(discriminant)) / (2 * a)
        if t_plus < 0:
            return np.Inf
        elif t_minus < 0:
            for index in range(3):
                P[index] = P0[index] + t_plus * d0[index]
            for index in range(3):
                n[index] = (P[index] - C[index])
            return t_plus
        else:
            for index in range(3):
                P[index] = P0[index] + t_minus * d0[index]
            for index in range(3):
                n[index] = (P[index] - C[index])
            return t_minus


@numba.jit(nopython=True)
def box_intersection(V, L, W, H, d0, P0, P, n):
    """
        This function computes the intersection point P between a ray and
        a box, and the normal n vector of the surface of the box at that point.
        It returns the t value of the line defined by the ray that
        corresponds to point P.
    """

    t_min = float32(0)
    t_max = np.Inf
    side = 0
    Ci = np.zeros(3, dtype=float32)
    ni = np.zeros(3, dtype=float32)
    n_min = np.zeros(3, dtype=float32)
    n_max = np.zeros(3, dtype=float32)
    P0Ci = np.zeros(3, dtype=float32)
    while ((side < 6) and (t_min <= t_max)):
        if (side == 0):
            for index in range(3):
                Ci[index] = V[index]
            mymath.cross_product(L, H, ni)
        elif (side == 1):
            for index in range(3):
                Ci[index] = V[index] + W[index]
            for index in range(3):
                ni[index] = - ni[index]
        elif (side == 2):
            for index in range(3):
                Ci[index] = V[index]
            mymath.cross_product(H, W, ni)
        elif (side == 3):
            for index in range(3):
                Ci[index] = V[index] + L[index]
            for index in range(3):
                ni[index] = - ni[index]
        elif (side == 4):
            for index in range(3):
                Ci[index] = V[index]
            mymath.cross_product(W, L, ni)
        elif (side == 5):
            for index in range(3):
                Ci[index] = V[index] + H[index]
            for index in range(3):
                ni[index] = - ni[index]

        ray_vector_sense = mymath.dot_product(d0, ni)

        for index in range(3):
            P0Ci[index] = P0[index] - Ci[index]
        ray_point_sense = mymath.dot_product(P0Ci, ni)

        if ray_vector_sense < 0:
            t_int = -ray_point_sense / ray_vector_sense
            if (t_int >= t_min):
                for index in range(3):
                    n_min[index] = ni[index]
                t_min = t_int
        elif ray_vector_sense > 0:
            t_int = -ray_point_sense / ray_vector_sense
            if (t_int <= t_max):
                for index in range(3):
                    n_max[index] = ni[index]
                t_max = t_int
        elif ray_point_sense > 0:
            t_min = t_max + 1
        side += 1
    if (t_min > t_max):
        return np.Inf
    else:
        if (t_min <= 1e-6):
            for index in range(3):
                P[index] = P0[index] + t_max * d0[index]
            for index in range(3):
                n[index] = n_max[index]
            return float32(t_max)
        else:
            for index in range(3):
                P[index] = P0[index] + t_min * d0[index]
            for index in range(3):
                n[index] = n_min[index]
            return float32(t_min)


@numba.jit(nopython=True)
def triangle_intersection(C, a, b, d0, P0, P, n):
    """
        This function computes the intersection point P between a ray and
        a triangle, and the normal n vector of the surface of the triangle at
        that point. It returns the t value of the line defined by the ray that
        corresponds to point P.
    """

    # Get the n normal vector to the plane
    mymath.cross_product(a, b, n)
    if (mymath.dot_product(n, d0) > 0):
        for index in range(3):
            n[index] = -n[index]

    # Ray parallel to plane
    if(abs(mymath.dot_product(n, d0)) < 1e-6):
        return np.Inf
    # Get the t value of the ray line at which the intersection happens
    P0C = np.zeros(3, dtype=float32)
    for index in range(3):
        P0C[index] = P0[index] - C[index]
    t = - mymath.dot_product(P0C, n) / mymath.dot_product(n, d0)
    # Intersection "behind" the point P0 at from wich the ray originated
    if (t < 0):
        return np.Inf
    # Get the point of intersection of the ray with the plane
    for index in range(3):
        P[index] = P0[index] + t * d0[index]
    # Check if the intersection point P lies within the triangle
    a_dot_a = mymath.dot_product(a, a)
    b_dot_b = mymath.dot_product(b, b)
    a_dot_b = mymath.dot_product(a, b)
    PC = np.zeros(3, dtype=float32)
    for index in range(3):
        PC[index] = P[index] - C[index]
    PC_dot_a = mymath.dot_product(PC, a)
    PC_dot_b = mymath.dot_product(PC, b)
    A = 1 / (a_dot_a * b_dot_b - a_dot_b ** 2)
    alpha = A * (b_dot_b * PC_dot_a - a_dot_b * PC_dot_b)
    beta = A * (- a_dot_b * PC_dot_a + a_dot_a * PC_dot_b)
    # P inside triangle, there is an intersection
    if (alpha > 0 and beta > 0 and alpha + beta <= 1):
        return t
    # P outside the triangle, there is no intersection
    else:
        return np.Inf


@numba.jit(nopython=True)
def ellipsoid_intersection(C, u, v, w, d0, P0, P, n):
    """
        This function computes the intersection point P between a ray and
        a ellipsoid, and the normal n vector of the surface of the elipsoid
        at that point. It returns the t value of the line defined by the ray
        that corresponds to point P.
    """

    M = np.zeros((3, 3), dtype=float32)
    for i in range(3):
        M[i, 0] = u[i]
        M[i, 1] = v[i]
        M[i, 2] = w[i]
    M_inv = np.zeros((3, 3), dtype=float32)
    mymath.matrix_inverse(M, M_inv)

    P0C = np.zeros(3, dtype=np.float32)
    for index in range(3):
        P0C[index] = P0[index] - C[index]

    M_inv_d0 = np.zeros(3, dtype=float32)
    mymath.matrix_vector_mul(M_inv, d0, M_inv_d0)

    M_inv_P0C = np.zeros(3, dtype=float32)
    mymath.matrix_vector_mul(M_inv, P0C, M_inv_P0C)

    a = mymath.dot_product(M_inv_d0, M_inv_d0)
    b = 2 * mymath.dot_product(M_inv_P0C, M_inv_d0)
    c = mymath.dot_product(M_inv_P0C, M_inv_P0C) - 1
    discriminant = b ** 2 - 4 * a * c
    if discriminant < 0:
        return np.Inf
    else:
        t_plus = (- b + np.sqrt(discriminant)) / (2 * a)
        t_minus = (- b - np.sqrt(discriminant)) / (2 * a)
        if t_plus < 0:
            return np.Inf
        elif t_minus < 0:
            for index in range(3):
                P[index] = P0[index] + t_plus * d0[index]
            # Get normal
            M_inv_T = np.zeros((3, 3), dtype=float32)
            mymath.matrix_transpose(M_inv, M_inv_T)
            PC = np.zeros(3, dtype=float32)
            for index in range(3):
                PC[index] = (P[index] - C[index])
            n_prime = np.zeros(3, dtype=float32)
            mymath.matrix_vector_mul(M_inv, PC, n_prime)
            mymath.matrix_vector_mul(M_inv_T, n_prime, n)

            return t_plus
        else:
            for index in range(3):
                P[index] = P0[index] + t_minus * d0[index]
            # Get normal
            M_inv_T = np.zeros((3, 3), dtype=float32)
            mymath.matrix_transpose(M_inv, M_inv_T)
            PC = np.zeros(3, dtype=float32)
            for index in range(3):
                PC[index] = (P[index] - C[index])
            n_prime = np.zeros(3, dtype=float32)
            mymath.matrix_vector_mul(M_inv, PC, n_prime)
            mymath.matrix_vector_mul(M_inv_T, n_prime, n)

            return t_minus


@numba.jit(nopython=True)
def OR_intersection(index_of_first, objects, d0, P0, medium_0, P_OR,
                    n_OR, t_medium_OR, boolean_op_num):
    """
        This function computes the intersection point P between a ray and
        a OR construct, and the normal n vector of the surface of the construct
        at that point. It returns the t value of the line defined by the ray
        that corresponds to point P.
    """

    t_tot = math.inf
    P = np.zeros(3, dtype=np.float32)
    n = np.zeros(3, dtype=np.float32)

    # Count number of operands
    OR_obj_i = index_of_first
    while (objects[OR_obj_i, 23] == 1):
        OR_obj_i += 1
    boolean_op_num[0] = OR_obj_i - index_of_first

    if (medium_0 == 0):
        t_medium_OR[0] = 0

        t_closest = math.inf
        for object_i in range(int(index_of_first),
                              int(index_of_first+boolean_op_num[0])):
            if (objects[object_i, 0] == 1):
                t = sphere_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14],
                    d0, P0, P, n)
            elif (objects[object_i, 0] == 2):
                t = box_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14:17],
                    objects[object_i, 17:20],
                    objects[object_i, 20:23],
                    d0, P0, P, n)
            elif (objects[object_i, 0] == 4):
                t = ellipsoid_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14:17],
                    objects[object_i, 17:20],
                    objects[object_i, 20:23],
                    d0, P0, P, n)
            if (t < t_closest):
                t_medium_OR[0] = 1000 + object_i
                t_closest = t
                for index in range(3):
                    P_OR[index] = P[index]
                for index in range(3):
                    n_OR[index] = n[index]
        t_tot = t_closest

    elif (medium_0 // 1000 == 1):
        t_medium_OR[0] = 0

        # inside_which array determines inside which of the object
        # ray_0 is. Even values indicate that the ray is outside the
        # object corresponding to the index. Odd values indicate that
        # the ray is inside the object. At the begining it can only
        # be inside one of the objects.
        inside_which = np.zeros(boolean_op_num[0], dtype=np.uint8)
        # ray_0 is in medium_0. Store this information in inside_which.
        inside_which[int(medium_0-1000-index_of_first)] += 1

        P_new = np.zeros(3, dtype=float32)
        epsilon = float32(1.0e-3)

        # For the first iteration, P_new is equal to P0
        P_new[:] = P0[:]

        t_tot = 0
        inside = True
        while inside:
            t_closest = math.inf
            index_of_closest = 0
            for object_i in range(int(index_of_first),
                                  int(index_of_first+boolean_op_num[0])):
                if (objects[object_i, 0] == 1):
                    t = sphere_intersection(
                        objects[object_i, 11:14],
                        objects[object_i, 14],
                        d0, P_new, P, n)
                elif (objects[object_i, 0] == 2):
                    t = box_intersection(
                        objects[object_i, 11:14],
                        objects[object_i, 14:17],
                        objects[object_i, 17:20],
                        objects[object_i, 20:23],
                        d0, P_new, P, n)
                elif (objects[object_i, 0] == 4):
                    t = ellipsoid_intersection(
                        objects[object_i, 11:14],
                        objects[object_i, 14:17],
                        objects[object_i, 17:20],
                        objects[object_i, 20:23],
                        d0, P_new, P, n)
                if (t < t_closest):
                    t_closest = t
                    index_of_closest = object_i
                    for index in range(3):
                        P_OR[index] = P[index]
                    for index in range(3):
                        n_OR[index] = n[index]

            # Analiticaly this if statement should always be false
            if (t_closest >= math.inf):
                inside = False
                # At this point t_tot should't be 0 (but sometimes it is)
                if (t_tot == 0):
                    t_tot = math.inf

            else:
                # Update inside_which
                inside_which[int(index_of_closest-index_of_first)] += 1
                # Add t value of last segment to the rest
                t_tot += t_closest

                for index in range(inside_which.shape[0]):
                    if (inside_which[index] == 1):
                        inside = True
                        break
                    inside = False

                norm_n_OR = math.sqrt(mymath.dot_product(n_OR, n_OR))

                # Get new origin point for ray (this origin point will be
                # positioned past the previous intersected surface).
                if (mymath.dot_product(d0, n_OR) > 0):
                    for index in range(3):
                        P_new[index] = P_OR[index] +\
                            epsilon * n_OR[index] / norm_n_OR
                else:
                    for index in range(3):
                        P_new[index] = P_OR[index] -\
                            epsilon * n_OR[index] / norm_n_OR
    return t_tot


@numba.jit(nopython=True)
def AND_intersection(index_of_first, objects, d0, P0, medium_0, P_AND,
                     n_AND, t_medium_AND, boolean_op_num):
    """
        This function computes the intersection point P between a ray and
        a AND construct, and the normal n vector of the surface of the
        construct at that point. It returns the t value of the line defined by
        the ray that corresponds to point P.
    """
    t_tot = math.inf
    P = np.zeros(3, dtype=np.float32)
    n = np.zeros(3, dtype=np.float32)

    # Count number of operands
    AND_obj_i = index_of_first
    while (objects[AND_obj_i, 23] == 2):
        AND_obj_i += 1
    boolean_op_num[0] = AND_obj_i - index_of_first

    if (medium_0 == 0):

        # inside_which array determines inside which of the object
        # ray_0 is. Even values indicate that the ray is outside the
        # object corresponding to the index. Odd values indicate that
        # the ray is inside the object. At the begining it can only
        # be inside one of the objects.
        inside_which = np.zeros(boolean_op_num[0], dtype=np.uint8)

        P_new = np.zeros(3, dtype=float32)
        epsilon = float32(1.0e-6)

        # For the first iteration, P_new is equal to P0
        P_new[:] = P0[:]

        t_tot = 0
        outside = True
        AND_possible = True
        while (outside and AND_possible):
            t_closest = math.inf
            index_of_closest = 0
            for object_i in range(int(index_of_first),
                                  int(index_of_first+boolean_op_num[0])):
                if (objects[object_i, 0] == 1):
                    t = sphere_intersection(
                        objects[object_i, 11:14],
                        objects[object_i, 14],
                        d0, P_new, P, n)
                elif (objects[object_i, 0] == 2):
                    t = box_intersection(
                        objects[object_i, 11:14],
                        objects[object_i, 14:17],
                        objects[object_i, 17:20],
                        objects[object_i, 20:23],
                        d0, P_new, P, n)
                elif (objects[object_i, 0] == 4):
                    t = ellipsoid_intersection(
                        objects[object_i, 11:14],
                        objects[object_i, 14:17],
                        objects[object_i, 17:20],
                        objects[object_i, 20:23],
                        d0, P_new, P, n)
                if (t < t_closest):
                    t_closest = t
                    index_of_closest = object_i
                    for index in range(3):
                        P_AND[index] = P[index]
                    for index in range(3):
                        n_AND[index] = n[index]

            if (t_closest >= math.inf):
                AND_possible = False
                t_tot = math.inf
                t_medium_AND[0] = 0

            else:
                # Update inside_which
                inside_which[int(index_of_closest-index_of_first)] += 1

                # Add t value of last segment to the rest
                t_tot += t_closest

                for index in range(inside_which.shape[0]):
                    if (inside_which[index] >= 2):
                        AND_possible = False
                        t_tot = math.inf
                        t_medium_AND[0] = 0
                        break
                    if (inside_which[index] % 2 != 1):
                        outside = True
                        break
                    outside = False
                    t_medium_AND[0] = 2

                norm_n_AND = math.sqrt(mymath.dot_product(n_AND, n_AND))

                # Get new origin point for ray (this origin point will be
                # positioned past the previous intersected surface).
                if (mymath.dot_product(d0, n_AND) > 0):
                    for index in range(3):
                        P_new[index] = P_AND[index] +\
                            epsilon * n_AND[index] / norm_n_AND
                else:
                    for index in range(3):
                        P_new[index] = P_AND[index] -\
                            epsilon * n_AND[index] / norm_n_AND

    elif (medium_0 == 2):
        t_medium_AND[0] = 0

        t_closest = math.inf
        for object_i in range(int(index_of_first),
                              int(index_of_first+boolean_op_num[0])):
            if (objects[object_i, 0] == 1):
                t = sphere_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14],
                    d0, P0, P, n)
            elif (objects[object_i, 0] == 2):
                t = box_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14:17],
                    objects[object_i, 17:20],
                    objects[object_i, 20:23],
                    d0, P0, P, n)
            elif (objects[object_i, 0] == 4):
                t = ellipsoid_intersection(
                    objects[object_i, 11:14],
                    objects[object_i, 14:17],
                    objects[object_i, 17:20],
                    objects[object_i, 20:23],
                    d0, P0, P, n)
            if (t < t_closest):
                t_closest = t
                for index in range(3):
                    P_AND[index] = P[index]
                for index in range(3):
                    n_AND[index] = n[index]
        t_tot = t_closest
    return t_tot
