
"""
.. module:: ray_casting.py
    :synopsis: It contains functions that compute the origin point and
    direction components of different rays.
.. moduleauthor :: Kereman Alonso (alonso.kerman5@gmail.com)
"""


import math
import numpy as np
import numba
from numpy import float32, pi, cos, sin

import modules.math_operations as mymath


@numba.jit(nopython=True)
def eye_ray(camera, alpha, beta, medium, ray):
    """
        This function computes the parameters of a ray that passes through
        the "eye" defined by the camera and the pixel in the screen with
        coordinates alpha and beta. The parameters are stored in ray.
    """
    # Pinhole camera
    if camera.shape[0] == 12:
        for i in range(3):
            ray[0, i] = camera[i] + camera[i+9]
            ray[1, i] = camera[i] + alpha * camera[i+3] + \
                beta * camera[i+6] - ray[0, i]
        ray[2, 0] = medium
    # Thin lens camera
    elif camera.shape[0] == 14:
        u_norm = math.sqrt(mymath.dot_product(camera[3:6], camera[3:6]))
        v_norm = math.sqrt(mymath.dot_product(camera[6:9], camera[6:9]))

        lens_vector = np.zeros(3, dtype=float32)
        u_component = 2 * np.random.random_sample() - 1
        v_component = 2 * np.random.random_sample() - 1
        for i in range(3):
            lens_vector[i] = u_component * camera[i+3] / u_norm +\
                v_component * camera[i+6] / v_norm
        lens_vector_norm = math.sqrt(
            mymath.dot_product(lens_vector, lens_vector))
        new_norm = camera[13] * np.random.random_sample()
        for i in range(3):
            lens_vector[i] = lens_vector[i] / lens_vector_norm * new_norm

        d_o = camera[12]
        d_i = math.sqrt(mymath.dot_product(camera[9:12], camera[9:12]))
        for i in range(3):
            ray[0, i] = camera[i] + camera[i+9] + lens_vector[i]
            ray[1, i] = (alpha * camera[i+3] + beta * camera[i+6] -
                         camera[i+9]) * d_o / d_i - lens_vector[i]
        ray[2, 0] = medium


@numba.jit(nopython=True)
def shadow_feeler(light_point, r, P, ray):
    """
        This function computes the parameters of a ray that originates at point
        P and passes through light_point. The parameters are stored in ray.
    """
    if r == 0:
        for i in range(3):
            ray[0, i] = P[i]
            ray[1, i] = light_point[i] - P[i]
        D = math.sqrt(mymath.dot_product(ray[1], ray[1]))
        for i in range(3):
            ray[1, i] = ray[1, i] / D
        ray[2, 0] = 0
        return float32(D)
    else:
        light_vector = np.zeros(3, dtype=float32)
        for i in range(3):
            light_vector[i] = 2 * np.random.random_sample() - 1
        light_vector_norm = math.sqrt(
            mymath.dot_product(light_vector, light_vector))
        for i in range(3):
            light_vector[i] = light_vector[i] / light_vector_norm * r

        for i in range(3):
            ray[0, i] = P[i]
            ray[1, i] = light_point[i] + light_vector[i] - P[i]
        D = math.sqrt(mymath.dot_product(ray[1], ray[1]))
        for i in range(3):
            ray[1, i] = ray[1, i] / D
        ray[2, 0] = 0
        return float32(D)


@numba.jit(nopython=True)
def reflected_ray(P, n, d0, medium, ray):
    """
        This function computes the reflection of a ray incident on a surface.
    """
    # Get a normal with the correct sense (towards incoming ray)
    n_prime = np.zeros(3, dtype=float32)
    if mymath.dot_product(n, d0) < 0:
        for index in range(3):
            n_prime[index] = n[index]
    else:
        for index in range(3):
            n_prime[index] = -n[index]
    # Get the normalization factor of n_prime
    norm_n_prime = math.sqrt(mymath.dot_product(n_prime, n_prime))
    # Normalize n_prime
    for index in range(3):
        n_prime[index] = n_prime[index] / norm_n_prime
    # Calculate the origin point of the reflected ray (epsilon distance
    # away from P point in n direction to ensure it is in the correct side
    # of the space divided by the surface)
    epsilon = float32(1.0e-3)
    P_epsilon = np.zeros(3, dtype=float32)
    for index in range(3):
        P_epsilon[index] = P[index] + epsilon * n_prime[index]

    norm_d0 = math.sqrt(mymath.dot_product(d0, d0))
    for i in range(3):
        ray[0, i] = P_epsilon[i]
        ray[1, i] = d0[i] / norm_d0 -\
            2 * mymath.dot_product(d0, n_prime) / norm_d0 *\
            n_prime[i]
    ray[2, 0] = medium


@numba.jit(nopython=True)
def glossy_reflected_ray(P, n, d0, ray_refl, medium, Se,
                         ray_glossy, cos_theta):
    """
        This function computes the glossy reflection of a ray incident on a
        surface.

    """
    # Get a normal with the correct sense (towards incoming ray)
    n_prime = np.zeros(3, dtype=float32)
    if mymath.dot_product(n, d0) < 0:
        for index in range(3):
            n_prime[index] = n[index]
    else:
        for index in range(3):
            n_prime[index] = -n[index]

    # Get the origin point and medium of the glossy reflection ray
    for i in range(3):
        ray_glossy[0, i] = ray_refl[0, i]
    ray_glossy[2, 0] = medium

    # Calculate the direction of the glossy reflection ray
    w = np.zeros(3, dtype=float32)
    v = np.zeros(3, dtype=float32)
    u = np.zeros(3, dtype=float32)

    w[:] = ray_refl[1, :]
    mymath.cross_product(w, d0, v)
    norm_v = math.sqrt(mymath.dot_product(v, v))
    v[:] = v[:]/norm_v
    mymath.cross_product(v, w, u)

    phi = 2 * pi * np.random.random_sample()
    cos_theta[0] = (1-np.random.random_sample()) ** (1/(Se+1))
    sin_theta = math.sqrt(1-cos_theta[0]**2)
    s_u = sin_theta * cos(phi)
    s_v = sin_theta * sin(phi)
    s_w = cos_theta[0]

    b = np.zeros(3, dtype=float32)
    b[:] = s_u * u[:] + s_v * v[:] + s_w * w[:]
    if mymath.dot_product(n_prime, b) < 0:
        ray_glossy[1, :] = - s_u * u[:] - s_v * v[:] + s_w * w[:]
    else:
        ray_glossy[1, :] = b[:]


@numba.jit(nopython=True)
def transmited_ray(P, n, d0, ni, nt, medium, ray):
    """
        This function computes the refraction of a ray incident on a surface,
        according to Snell's Law of Refraction
    """

    epsilon = float32(1.0e-3)
    P_epsilon = np.zeros(3, dtype=float32)
    n_ratio = ni / nt
    norm_n = math.sqrt(mymath.dot_product(n, n))
    norm_d0 = math.sqrt(mymath.dot_product(d0, d0))
    # Get the dot product
    d0_dot_n = mymath.dot_product(d0, n) / (norm_n * norm_d0)
    # Depending on how the normal n is defined, t_dot_n should be equal to
    # cos(theta_t) or - cos(theta_t)
    if (d0_dot_n <= 0):
        t_dot_n = - math.sqrt(1 - n_ratio ** 2 * (1 - d0_dot_n ** 2))
        for index in range(3):
            P_epsilon[index] = P[index] - epsilon * n[index] / norm_n
    else:
        t_dot_n = math.sqrt(1 - n_ratio ** 2 * (1 - d0_dot_n ** 2))
        for index in range(3):
            P_epsilon[index] = P[index] + epsilon * n[index] / norm_n
    for i in range(3):
        ray[0, i] = P_epsilon[i]
        ray[1, i] = (t_dot_n - n_ratio * d0_dot_n) * n[i] /\
            norm_n + n_ratio * d0[i] / norm_d0
    ray[2, 0] = medium
