
"""
.. module:: renderGPU.py
    :synopsis: It contains the rendering function that calculates the color of
    each pixel using ray tracing techniques.

.. moduleauthor :: Kereman Alonso (alonso.kerman5@gmail.com)
"""


import numpy as np
from numpy import float32, uint8, int32
import numba
import math

import modules.math_operations as mymath
import modules.ray_casting as ray_casting
import modules.color as color
import modules.intersections as intersections


@numba.jit(nopython=True)
def compute_radiance(pixel_data, corner_pixel_data, objects, camera, lights,
                     ambient, air,  i, j, d, sampling, order):
    """
        This function renders the color of a pixel.
    """

    horz_scaling = pixel_data.shape[1] / 2.0
    vert_scaling = pixel_data.shape[0] / 2.0
    alpha = (j - horz_scaling) / horz_scaling
    beta = -(i - vert_scaling) / vert_scaling

    color_total_RGB = np.zeros(3, dtype=float32)

    if (sampling == 'one_per_pixel'):
        # Cast ray_v that originates in the camera and goes through the pixel
        # into the scene
        alpha += 0.5 / horz_scaling
        beta += 0.5 / horz_scaling
        ray_v = np.zeros((3, 3), dtype=np.float32)
        medium = 0
        ray_casting.eye_ray(camera, alpha, beta, medium, ray_v)

        ni = float32(1)
        color_total_linearRGB = np.zeros(3, dtype=float32)

        # Use recursion to compute global illumination
        ray_trace(objects, lights, ambient, air, ray_v, ni, d,
                  color_total_linearRGB)

        color.linearRGB_to_sRGB(color_total_linearRGB, color_total_RGB)

    elif (sampling == 'sup_uniform'):
        uniform_supersampling(camera, objects, lights, ambient, air,
                              alpha, beta, horz_scaling, vert_scaling, d,
                              order, color_total_RGB)

    elif (sampling == 'sup_jittered'):
        jittered_supersampling(camera, objects, lights, ambient, air,
                               alpha, beta, horz_scaling, vert_scaling, d,
                               order, color_total_RGB)

    elif (sampling == 'corners'):
        beta += 1 / vert_scaling

        ray_v = np.zeros((3, 3), dtype=np.float32)
        medium = 0
        ray_casting.eye_ray(camera, alpha, beta, medium, ray_v)

        ni = float32(1)
        color_total_linearRGB = np.zeros(3, dtype=float32)

        # Use recursion to compute global illumination
        ray_trace(objects, lights, ambient, air, ray_v, ni, d,
                  color_total_linearRGB)

        for RGB in range(3):
            corner_pixel_data[i, j, RGB] = color_total_linearRGB[RGB]
        return

    elif (sampling == 'sup_adaptative'):
        color_total_linearRGB = np.zeros(3, dtype=float32)
        corner_colors = np.zeros((4, 3), dtype=float32)

        for RGB in range(3):
            corner_colors[0, RGB] = corner_pixel_data[i+1, j, RGB]
            corner_colors[1, RGB] = corner_pixel_data[i+1, j+1, RGB]
            corner_colors[2, RGB] = corner_pixel_data[i, j, RGB]
            corner_colors[3, RGB] = corner_pixel_data[i, j+1, RGB]

        adaptative_supersampling(camera, objects, lights, ambient, air,
                                 alpha, beta, horz_scaling,
                                 vert_scaling, d, order, corner_colors,
                                 color_total_linearRGB)

        color.linearRGB_to_sRGB(color_total_linearRGB, color_total_RGB)

    for RGB in range(3):
        int_RGB = int32(color_total_RGB[RGB] * 255)
        if (int_RGB <= 255):
            pixel_data[i, j, RGB] = uint8(int_RGB)
        else:
            pixel_data[i, j, RGB] = uint8(255)


@numba.jit(nopython=True)
def uniform_supersampling(camera, objects, lights, ambient, air,
                          alpha_pix, beta_pix, horz_scaling,
                          vert_scaling, d, order, color_total_RGB):
    """
        This function renders the color of a pixel aplying uniform
        supersampling.
    """

    ray_v = np.zeros((3, 3), dtype=np.float32)
    medium = 0
    ni = float32(1)
    num_of_samples = order ** 2

    color_linear = np.zeros(3, dtype=float32)
    color_total_linear = np.zeros(3, dtype=float32)

    for i in range(order):
        for j in range(order):
            alpha = alpha_pix + (0.5 + i) / (order * horz_scaling)
            beta = beta_pix + (0.5 + j) / (order * vert_scaling)
            ray_casting.eye_ray(camera, alpha, beta, medium, ray_v)
            ray_trace(objects, lights, ambient, air, ray_v, ni, d,
                      color_linear)

            for RGB in range(3):
                color_total_linear[RGB] +=\
                    color_linear[RGB] / num_of_samples

    color.linearRGB_to_sRGB(color_total_linear, color_total_RGB)


@numba.jit(nopython=True)
def jittered_supersampling(camera, objects, lights, ambient, air,
                           alpha_pix, beta_pix, horz_scaling,
                           vert_scaling, d, order, color_total_RGB):
    """
        This function renders the color of a pixel aplying jittered
        supersampling.
    """

    ray_v = np.zeros((3, 3), dtype=np.float32)
    medium = 0
    ni = float32(1)
    num_of_samples = order ** 2

    color_linear = np.zeros(3, dtype=float32)
    color_total_linear = np.zeros(3, dtype=float32)

    for i in range(order):
        for j in range(order):
            alpha = alpha_pix + (np.random.random_sample() + i) /\
                (order * horz_scaling)
            beta = beta_pix + (np.random.random_sample() + j) /\
                (order * vert_scaling)
            ray_casting.eye_ray(camera, alpha, beta, medium, ray_v)
            ray_trace(objects, lights, ambient, air, ray_v, ni, d,
                      color_linear)

            for RGB in range(3):
                color_total_linear[RGB] +=\
                    color_linear[RGB] / num_of_samples

    color.linearRGB_to_sRGB(color_total_linear, color_total_RGB)


@numba.jit(nopython=True)
def adaptative_supersampling(camera, objects, lights, ambient, air,
                             alpha_pix, beta_pix, horz_scaling,
                             vert_scaling, d, order, corner_colors,
                             color_total_linear):
    """
        This function renders the color of a pixel aplying adaptive
        supersampling.
    """
    if (order < 0):
        need_supersampling = False
    else:
        need_supersampling = False
        for i in range(3):
            for j in range(i+1, 4):
                for RGB in range(3):
                    if (abs(corner_colors[i, RGB] -
                            corner_colors[j, RGB]) > 0.05):
                        need_supersampling = True

    if (need_supersampling is True):
        ray_v = np.zeros((3, 3), dtype=np.float32)
        medium = 0
        ni = float32(1)

        # Get the colors of 5 additional points
        lower_centre = np.zeros(3, dtype=float32)
        alpha = alpha_pix + 1 / (2 * horz_scaling)
        beta = beta_pix + 0 / (2 * vert_scaling)
        ray_casting.eye_ray(camera, alpha, beta, medium, ray_v)
        ray_trace(objects, lights, ambient, air, ray_v, ni, d,
                  lower_centre)

        centre_left = np.zeros(3, dtype=float32)
        alpha = alpha_pix + 0 / (2 * horz_scaling)
        beta = beta_pix + 1 / (2 * vert_scaling)
        ray_casting.eye_ray(camera, alpha, beta, medium, ray_v)
        ray_trace(objects, lights, ambient, air, ray_v, ni, d,
                  centre_left)

        centre_centre = np.zeros(3, dtype=float32)
        alpha = alpha_pix + 1 / (2 * horz_scaling)
        beta = beta_pix + 1 / (2 * vert_scaling)
        ray_casting.eye_ray(camera, alpha, beta, medium, ray_v)
        ray_trace(objects, lights, ambient, air, ray_v, ni, d,
                  centre_centre)

        centre_right = np.zeros(3, dtype=float32)
        alpha = alpha_pix + 2 / (2 * horz_scaling)
        beta = beta_pix + 1 / (2 * vert_scaling)
        ray_casting.eye_ray(camera, alpha, beta, medium, ray_v)
        ray_trace(objects, lights, ambient, air, ray_v, ni, d,
                  centre_right)

        upper_centre = np.zeros(3, dtype=float32)
        alpha = alpha_pix + 1 / (2 * horz_scaling)
        beta = beta_pix + 2 / (2 * vert_scaling)
        ray_casting.eye_ray(camera, alpha, beta, medium, ray_v)
        ray_trace(objects, lights, ambient, air, ray_v, ni, d,
                  upper_centre)

        # Call addaptative supersampling for the four subdivisions of the
        # pixel. Store the colors in corner_colors array.
        new_horz_scaling = 2 * horz_scaling
        new_vert_scaling = 2 * vert_scaling

        # Lower left corner color.
        sub_total_color = np.zeros(3, dtype=float32)

        sub_corner_colors = np.zeros((4, 3), dtype=float32)
        sub_corner_colors[0, :] = corner_colors[0, :]
        sub_corner_colors[1, :] = lower_centre[:]
        sub_corner_colors[2, :] = centre_left[:]
        sub_corner_colors[3, :] = centre_centre[:]

        alpha = alpha_pix
        beta = beta_pix
        adaptative_supersampling(camera, objects, lights, ambient, air,
                                 alpha, beta, new_horz_scaling,
                                 new_vert_scaling, d, order - 1,
                                 sub_corner_colors,
                                 sub_total_color)
        corner_colors[0, :] = sub_total_color[:]

        # Lower right corner color.
        sub_corner_colors[0, :] = lower_centre[:]
        sub_corner_colors[1, :] = corner_colors[1, :]
        sub_corner_colors[2, :] = centre_centre[:]
        sub_corner_colors[3, :] = centre_right[:]

        alpha = alpha_pix + 1 / (2 * horz_scaling)
        beta = beta_pix + 0 / (2 * vert_scaling)
        adaptative_supersampling(camera, objects, lights, ambient, air,
                                 alpha, beta, new_horz_scaling,
                                 new_vert_scaling, d, order-1,
                                 sub_corner_colors,
                                 sub_total_color)
        corner_colors[1, :] = sub_total_color[:]

        # Upper left corner color.
        sub_corner_colors[0, :] = centre_left[:]
        sub_corner_colors[1, :] = centre_centre[:]
        sub_corner_colors[2, :] = corner_colors[2, :]
        sub_corner_colors[3, :] = upper_centre[:]

        alpha = alpha_pix + 0 / (2 * horz_scaling)
        beta = beta_pix + 1 / (2 * vert_scaling)
        adaptative_supersampling(camera, objects, lights, ambient, air,
                                 alpha, beta, new_horz_scaling,
                                 new_vert_scaling, d, order-1,
                                 sub_corner_colors,
                                 sub_total_color)
        corner_colors[2, :] = sub_total_color[:]

        # Upper right corner color
        sub_corner_colors[0, :] = centre_centre[:]
        sub_corner_colors[1, :] = centre_right[:]
        sub_corner_colors[2, :] = upper_centre[:]
        sub_corner_colors[3, :] = corner_colors[3, :]

        alpha = alpha_pix + 1 / (2 * horz_scaling)
        beta = beta_pix + 1 / (2 * vert_scaling)
        adaptative_supersampling(camera, objects, lights, ambient, air,
                                 alpha, beta, new_horz_scaling,
                                 new_vert_scaling, d, order-1,
                                 sub_corner_colors,
                                 sub_total_color)
        corner_colors[3, :] = sub_total_color[:]

    for RGB in range(3):
        color_total_linear[RGB] =\
            (corner_colors[0, RGB] + corner_colors[1, RGB] +
             corner_colors[2, RGB] + corner_colors[3, RGB]) / 4


@numba.jit(nopython=True)
def ray_trace(objects, lights, ambient, air, ray_v, ni, d,
              color_total_linear):
    """
        This function renders the color of a light ray.
    """
    # Initialize values of color_total_linear to 0
    color_total_linear[:] = np.zeros(3, dtype=float32)[:]

    # If maximum depth of recursion has been reached, return
    if d < 0:
        return

    # Get the index of the first object that the ray intersects.
    # Get the point P_closest where the intersection happens.
    # Get the normal vector n_closest to the objects surface at P_closest.
    # Get the distance to the object.
    P_closest = np.zeros(3, dtype=np.float32)
    n_closest = np.zeros(3, dtype=np.float32)
    distance = np.zeros(1, dtype=np.float32)
    t_medium = np.zeros(1, dtype=np.float32)
    cos_theta = np.zeros(1, dtype=np.float32)
    index_of_closest = intersections.intersection(objects,
                                                  ray_v, P_closest,
                                                  n_closest, distance,
                                                  t_medium)
    # If ray_v has not intersected with an object, return
    if (index_of_closest == - 1):
        return

    # Get nt assuming that all objects in the scene are surrounded by air,
    # even when two objects apear to be in contact.

    if (ni - 1 < 1e-6):
        nt = math.sqrt(objects[index_of_closest, 9] *
                       objects[index_of_closest, 10])
    else:
        nt = float32(1)

    # Get the reflection coefficient R and transmission coefficient T
    R = reflection_coefficient(ray_v, n_closest, ni, nt)
    T = 1 - R
    # Multiply R and T by the specular reflection coefficient
    R = R * objects[index_of_closest, 4]
    T = T * objects[index_of_closest, 4]

    # If ray_v has intersected with an object, compute local radiance
    # component
    local_illumination(objects, lights, ambient, ray_v[1],
                       P_closest, n_closest, index_of_closest,
                       color_total_linear)

    # Reflected ray radiance component
    if (R > 1e-5):
        reflection_num = 1

        ray_refl = np.zeros((3, 3), dtype=float32)
        r_medium = ray_v[2, 0]
        ray_casting.reflected_ray(P_closest, n_closest, ray_v[1], r_medium,
                                  ray_refl)
        if reflection_num == 1:
            color_refl = np.zeros(3, dtype=float32)
            ray_trace(objects, lights, ambient, air, ray_refl, ni, d-1,
                      color_refl)
            for RGB in range(3):
                color_total_linear[RGB] += R * color_refl[RGB]
        else:
            Se = objects[index_of_closest, 5]
            ray_glossy = np.zeros((3, 3), dtype=float32)
            for reflection_i in range(reflection_num):
                ray_casting.glossy_reflected_ray(
                    P_closest, n_closest, ray_v[1], ray_refl, r_medium, Se,
                    ray_glossy, cos_theta)
                color_refl = np.zeros(3, dtype=float32)
                ray_trace(objects, lights, ambient, air, ray_glossy, ni, d-1,
                          color_refl)
                for RGB in range(3):
                    color_total_linear[RGB] += R * color_refl[RGB] *\
                        cos_theta[0] ** Se / reflection_num

    # Transmited ray radiance component
    if (T > 1e-5):
        ray_trans = np.zeros((3, 3), dtype=float32)
        ray_casting.transmited_ray(P_closest, n_closest,
                                   ray_v[1], ni, nt, t_medium[0], ray_trans)
        color_trans = np.zeros(3, dtype=float32)
        ray_trace(objects, lights, ambient, air, ray_trans, nt, d-1,
                  color_trans)
        for RGB in range(3):
            color_total_linear[RGB] += T * color_trans[RGB]

    # Apply the attenuation factors corresponding to the medium through which
    # ray_v is travelling.
    color_total_RGB = np.zeros(3, dtype=float32)
    color.linearRGB_to_sRGB(color_total_linear, color_total_RGB)
    if (ni - 1 < 1e-6):
        for RGB in range(3):
            color_total_RGB[RGB] = color_total_RGB[RGB] *\
                air[2+RGB] ** distance[0]
    else:
        for RGB in range(3):
            color_total_RGB[RGB] = color_total_RGB[RGB] *\
                objects[index_of_closest, 6+RGB] ** distance[0]

    color.sRGB_to_linearRGB(color_total_RGB, color_total_linear)


@numba.jit(nopython=True)
def local_illumination(objects, lights, ambient, d0, P_closest,
                       n_closest, index_of_closest, total_radiance_RGB):
    """
        This function calculates the diffuse, specular and ambient light
        components and ads them to get the local illumination at point
        P_closest.
    """

    # Initialize values of total_radiance_RGB to 0
    total_radiance_RGB[:] = np.zeros(3, dtype=float32)[:]

    # If ray_v has intersected with an object, cast shadow feelers from
    # P_closest to each light source. If a light source is reached, add the
    # radiance that reaches the eye through the pixel, and color the pixel.
    P = np.zeros(3, dtype=np.float32)
    n = np.zeros(3, dtype=np.float32)

    linearRGB = np.zeros(3, dtype=float32)

    norm_n = math.sqrt(mymath.dot_product(n_closest, n_closest))
    epsilon = float32(1.0e-5)
    P_epsilon = np.zeros(3, dtype=float32)

    for index in range(3):
        P_epsilon[index] = P_closest[index] + epsilon * n_closest[index] /\
            norm_n

    ray_l = np.zeros((3, 3), dtype=float32)
    ray_rl = np.zeros((3, 3), dtype=float32)
    t_medium = np.zeros(1, dtype=float32)
    boolean_op_num = np.zeros(1, dtype=int32)
    shadow_num = 1
    for light in range(len(lights)):
        for shadow_i in range(shadow_num):
            if shadow_num == 1:
                r = 0
            else:
                r = lights[light, 6]
            D = ray_casting.shadow_feeler(
                lights[light, 0:3], r,  P_epsilon, ray_l)
            shadow = False

            object_i = 0
            while (object_i < len(objects)):
                if (objects[object_i, 23] == 0):
                    if (objects[object_i, 0] == 1):
                        t = intersections.sphere_intersection(
                            objects[object_i, 11:14],
                            objects[object_i, 14],
                            ray_l[1], ray_l[0], P, n)
                    elif (objects[object_i, 0] == 2):
                        t = intersections.box_intersection(
                            objects[object_i, 11:14],
                            objects[object_i, 14:17],
                            objects[object_i, 17:20],
                            objects[object_i, 20:23],
                            ray_l[1], ray_l[0], P, n)
                    elif (objects[object_i, 0] == 3):
                        t = intersections.triangle_intersection(
                            objects[object_i, 11:14],
                            objects[object_i, 14:17],
                            objects[object_i, 17:20],
                            ray_l[1], ray_l[0], P, n)
                    elif (objects[object_i, 0] == 4):
                        t = intersections.ellipsoid_intersection(
                            objects[object_i, 11:14],
                            objects[object_i, 14:17],
                            objects[object_i, 17:20],
                            objects[object_i, 20:23],
                            ray_l[1], ray_l[0], P, n)
                elif (objects[object_i, 23] == 1):
                    t = intersections.OR_intersection(object_i, objects,
                                                      ray_l[1],
                                                      ray_l[0],
                                                      ray_l[2, 0], P, n,
                                                      t_medium, boolean_op_num)
                    object_i += boolean_op_num[0] - 1
                elif (objects[object_i, 23] == 2):
                    t = intersections.AND_intersection(object_i, objects,
                                                       ray_l[1],
                                                       ray_l[0],
                                                       ray_l[2, 0], P, n,
                                                       t_medium,
                                                       boolean_op_num)
                    object_i += boolean_op_num[0] - 1
                if (t < D):
                    shadow = True
                    break
                object_i += 1

            if (shadow is False):

                # Diffuse reflection component
                nl = abs(mymath.dot_product(n_closest, ray_l[1]) / norm_n)
                for RGB in range(3):
                    linearRGB[RGB] = objects[index_of_closest, 1+RGB]\
                        * nl * lights[light, 3+RGB]
                # # Add colors in linear RGB
                for RGB in range(3):
                    total_radiance_RGB[RGB] += linearRGB[RGB] / shadow_num

                # Specular reflection component
                ray_casting.reflected_ray(P_closest, n_closest, -ray_l[1],
                                          ray_l[2, 0], ray_rl)
                rl_dot_d0 = mymath.dot_product(ray_rl[1], d0)
                if (rl_dot_d0 < 0):
                    norm_d0 = math.sqrt(mymath.dot_product(d0, d0))
                    rl_dot_d0 = rl_dot_d0 / norm_d0
                    Sc = objects[index_of_closest, 4]
                    Se = objects[index_of_closest, 5]
                    for RGB in range(3):
                        linearRGB[RGB] = Sc * abs(rl_dot_d0) ** Se *\
                            lights[light, 3+RGB]
                    # # Add colors in linear RGB
                    for RGB in range(3):
                        total_radiance_RGB[RGB] += linearRGB[RGB] / shadow_num

                # Ambient light component
                for RGB in range(3):
                    linearRGB[RGB] = objects[
                        index_of_closest, 1+RGB] * ambient[RGB]
                # # Add colors in linear RGB
                for RGB in range(3):
                    total_radiance_RGB[RGB] += linearRGB[RGB] / shadow_num


@numba.jit(nopython=True)
def reflection_coefficient(ray, n, ni, nt):
    """
        This function calculates the reflection coefficient of a surface by
        applying Fresnel equations. The ratio of the magnetic permeabilities
        of the two mediums is considered to be 1.
    """

    if (nt >= math.inf):
        return float32(1)
    n_ratio = ni / nt
    # Get cos(theta_i). cos(theta_i) should always be positive, if the normal
    # vector is chosen correctly. To take into acount that the given
    # normal vector n might be the inverse of what it should be, the absolute
    # value is computed.
    cos_theta_i = abs(mymath.dot_product(ray[1], n) /
                      (math.sqrt(mymath.dot_product(ray[1], ray[1])) *
                       math.sqrt(mymath.dot_product(n, n))))

    # Check if total internal reflaction happens. If it does return R = 1,
    # if not, get the value of cos(theta_t).
    discriminator = 1 - n_ratio ** 2 * (1 - cos_theta_i ** 2)
    if (discriminator <= 0):
        return float32(1)
    cos_theta_t = math.sqrt(discriminator)

    # For light polarized in a direction perpendicular to the plane of
    # incidence, get the ratio between the amplitude of the incident light
    # wave's electric field and the amplitude of the reflected light wave's
    # electric field.
    r_perp = (n_ratio * cos_theta_i - cos_theta_t) /\
        (n_ratio * cos_theta_i + cos_theta_t)

    # Get the same ratio for light polarized in a perpendicular direction to
    # the plane
    r_parallel = (cos_theta_i - n_ratio * cos_theta_t) /\
        (cos_theta_i + n_ratio * cos_theta_t)

    # Get the reflection coefficient R
    return float32(1.0 / 2 * (r_perp ** 2 + r_parallel ** 2))
